#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 6.36
#pragma ModuleName		= SpectraImageProfile
#pragma version			= 4.10
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

static Constant kVersion = 4.10
static StrConstant kVersionDate = "03/2023"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21549																// the project node on IgorExchange
static StrConstant ksShortTitle = "Profile Generator"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//
//	SpectraProfile is a environment to create 1D profiles from 2D data sets in horizontal or vertical
//	direction. Various modifications to the 2D data itself can be applied. Some functionality relies on
//	properties of the data used (such as scaling and column labels) and can be changed below.
//	
// 2021-02-16 - ver. 3.60:	Initial public release.
// 2021-07-22 - ver. 3.70:	Slightly adjusted the names in the main menu entries.
// 							Fixed wrong wave labels (row vs. col) for the split tool.
//							Added scaled positions to the wave name for the split tool.
//							Added keyboard shortcuts for Normalize Profile and Snap to Grid.
//							Added Accept & Quit button.
//							Added optional logging of data edits in the history.
//							Added control to set the smoothing factor for the align feature.
//							Moved all controls into an external panel.
//							Now the live profile under the mouse cursor gets normalized as well.
//							Fixed the way-too-big smoothing value when using the align feature.
//							Data units get displayed in the graph now.
// 2021-07-24 - ver. 3.71:	Improvements of matrix calculations thanks to suggestions by A.G. (WaveMetrics).
//							Added image contrast controls.
// 2021-10-26 - ver. 3.72:	When combining waves the source waves are recorded in the combined wave's notes now.
//							Added sub-step shifting to the Align feature.
// 2021-12-23 - ver. 3.80:	Made sure that sub-step alignment does not throw errors if the fit fails.
//							Fixed wrong panel size on windows with big fonts.
//							Align features with cursors present was not working correctly in some cases.
//							Added Align Edges feature which aligns features in the derivative.
//							Image profile now works with 3D data as well, either on the current layer or the merged image.
// 2022-03-23 - ver. 3.90:	Fixed error message when QuickProfile and QuickSplit are run with a folder selected.
//							Reduced ROI expansion via keyboard to single steps (unless snap-to-grid is activated).
//							Steps for repositioning and width changes via keyboard are now adapted to the data size.
//							Initial center value will be aligned to the grid for waves with less than 100 rows / columns, and honors zero.
//							Added Delete button to delete individual layers from 3D data.
//							Integration ranges are now indicated in the output name with 'to' instead of a hyphen/minus to avoid confusion.
//							Now multiple instances of the tool can be opened at the same time.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Axis split between image and profile can now be dragged to a new position.
// 2022-10-28 - ver. 4.00:	Added support for data units.
//							Fixed errors when selecting data folders instead of waves or when nothing is selected.
//							Now 3D volumes can be rotated around the three principal axes.
//							Axes are now reversed according to the sign of the delta-scaling in the respective direction.
//							Many tweaks and bug-fixes added.
//							---- Many thanks to Oleg Dyachok from Uppsala University for suggesting and testing the changes below ----
//							Holding both shift + control / command while using the arrow keys will change the center / width by exactly one unit.
//							Combining and stacking waves writes the input wave names as dimension labels.
//							Added display of multiple profiles generated from all layers of a 3D volume.
//							Added support for displaying dimension labels in the Image Profile tool.
//							Added color scale legend to the image.
// 2023-01-07 - ver. 4.05:	Added right-click menu to the Data Browser.
//							Added data-binning functionality.
//							Multi-profiles are now saved as well.
//							Dimension labels are rotated together with the data.
//							Added a 'Save as' button to save the input data under a new name (after edits etc.).
// 2023-03-19 - ver. 4.10:	Notes are now written into the output when splitting waves into rows / columns / layers.
//							Sub-range for integration can now be specified for Quick-Profile function.
//							Also, holding shift while invoking the menu summons a user interface to select the integration range.
//							Changed the meaning of the 'dim' parameter for the Quick Profile and Quick Split functions to mean the dimension to be integrated / split.
//							Fixed bug: The buttons of the inset panel could appear on different windows in some circumstances.
//							Fixed bug: Closing one instance broke updates in other instances.
//							Fixed bug: Some buttons were not sized properly for Mac.
//________________________________________________________________________________________________


//-------------------------------- functions for user code ---------------------------------------
static StrConstant kUserInfoBtnName = "Ring Current"											// the title of the user button

static Function Profile_ActivateUserInfoButton(inwave)											// called to check whether or not the data fulfills conditions for measurement normalization
	Wave inwave
	Variable activate = 0
	// +++++++++++++++++++++++++++ conditions for activation +++++++++++++++++++++++++++++++++++++
	activate = StringMatch(GetDimLabel(inwave,1,0),"*mA*")										// check if dim-labels contain "mA" values
	activate = StringMatch(note(inwave),"*ring current normalized*")? 0 : activate				// already normalized?
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return activate																				// return 1 to activate button
End

static Function Profile_ExecuteUserInfoButton(inwave)											// normalization of the data to measurement info (when the button is pressed)
	Wave inwave
	Variable valAvg = 0
	// ++++++++++++++++++++++ ring-current normalization procedure +++++++++++++++++++++++++++++++
	Variable i, value, dim = WaveDims(inwave)
	for (i = 0; i < DimSize(inwave,dim-1); i += 1)
		String Currlabel = GetDimLabel(inwave,1,i)												// get the information from the column label
		sscanf Currlabel, "%s (%f mA)", Currlabel, value										// search for ring current info in the form "time (XXX mA)"
		if (value != 0)
			if (dim == 3)
				inwave[][][i] /= value
			else
				inwave[][i] /= value
			endif
		endif
		valAvg += value																			// add the values for intensity re-normalization later
	endfor
	valAvg /= i+1
	Note inwave, "ring current normalized"														// write into note that the normalization has been done
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return valAvg																				// return the intensity normalization factor which may be used restore the average height
End
//------------------------------------------------------------------------------------------------

static StrConstant kWorkingDir	 = "Profile_Generator"											// the name for the working folder
static StrConstant kSettingsFile = "Spectra Tools settings.dat"									// just for keyboard shortcuts for now
static StrConstant kMultiProfileColor = "PastelsMap"											// color map name for multi-profile coloring
static Constant	kNormSmooth	 = 2		// profile normalization: integer factor for smoothing the profile before division (0 = no smoothing)
static Constant	kAlignSmooth = 2		// only for backwards compatibility - align features: smoothing factor in percent of the wave's x/y dimension (e.g., 2 => points * 0.02)
static Constant kMaxMultiProfiles = 101	// upper limit for the number of simultaneously displayed profiles of 3D volumes
static Constant kShowLabelDisplay = 1	// enable to show dimension-label annotation

Menu "Spectra Tools"
	SpectraImageProfile#MenuEntry("start_GUI")	,/Q, CreateBrowser; SpectraImageProfile#LoadWrapper()
	"-"
	SpectraImageProfile#MenuEntry("append_2D")	,/Q, CreateBrowser; CombineDataSets()
	SpectraImageProfile#MenuEntry("stack_2D")	,/Q, CreateBrowser; CombineDataSets(stack3D=1)
	SpectraImageProfile#MenuEntry("bin_data")	,/Q, CreateBrowser; QuickDataBinWrapper()
	"-"
	SpectraImageProfile#MenuEntry("profile_H")	,/Q, CreateBrowser; QuickMenuWrapper(1,"QProfile")
	SpectraImageProfile#MenuEntry("profile_V")	,/Q, CreateBrowser; QuickMenuWrapper(0,"QProfile")
	SpectraImageProfile#MenuEntry("profile_L")	,/Q, CreateBrowser; QuickMenuWrapper(2,"QProfile")
	"-"
	SpectraImageProfile#MenuEntry("split_H")	,/Q, CreateBrowser; QuickMenuWrapper(1,"QImageSplit")
	SpectraImageProfile#MenuEntry("split_V")	,/Q, CreateBrowser; QuickMenuWrapper(0,"QImageSplit")
	SpectraImageProfile#MenuEntry("split_L")	,/Q, CreateBrowser; QuickMenuWrapper(2,"QImageSplit")
End

#if (IgorVersion() >= 9.0)
Menu "DataBrowserObjectsPopup"
	"Append 1D,2D"				,/Q, CombineDataSets()
	"Stack 2D->3D"				,/Q, CombineDataSets(stack3D=1)
	"Average 2D,3D"				,/Q, QuickDataBinWrapper()
	"-"
	"Profile - [Horizontal]"	,/Q, QuickMenuWrapper(1,"QProfile")
	"Profile - [Vertical]"		,/Q, QuickMenuWrapper(0,"QProfile")
	"Merge 3D - [Layers]"		,/Q, QuickMenuWrapper(2,"QProfile")
	"-"
	"Split Data - [Horizontal]"	,/Q, QuickMenuWrapper(1,"QImageSplit")
	"Split Data - [Vertical]"	,/Q, QuickMenuWrapper(0,"QImageSplit")
	"Split Data - [Layers]"		,/Q, QuickMenuWrapper(2,"QImageSplit")
	"-"
End
#endif

//################################################################################################

static Function/S MenuEntry(which)
	String which
	Variable fileID

	String read = "", menuStr = "", searchStr = ""
	StrSwitch (which)
		case "start_GUI":
			menuStr = "Image Profile ..."
			searchStr = "start image profile"
		break
		case "append_2D":
			menuStr = "Append 1D,2D Datasets into 2D Data"
			searchStr = "append 2D data"
		break
		case "stack_2D":
			menuStr = "Stack 2D Datasets into 3D Data"
			searchStr = "stack 2D data"
		break
		case "bin_data":
			menuStr = "Average 2D|3D Data (Binning)"
			searchStr = "quick data binning"
		break
		case "profile_H":
			menuStr = "Profile of 2D|3D Data - [Horizontal]"
			searchStr = "quick horizontal profile"
		break
		case "profile_V":
			menuStr = "Profile of 2D|3D Data - [Vertical]"
			searchStr = "quick vertical profile"
		break
		case "profile_L":
			menuStr = "2D Image from 3D Data - [Layers]"
			searchStr = "quick layer profile"
		break
		case "split_H":
			menuStr = "Split Up 2D|3D Data - [Horizontal]"
			searchStr = "data split horizontal"
		break
		case "split_V":
			menuStr = "Split Up 2D|3D Data - [Vertical]"
			searchStr = "data split vertical"
		break
		case "split_L":
			menuStr = "Split Up 3D Data - [Layers]"
			searchStr = "data split layers"
		break
		default:
			return ""
	EndSwitch
	
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFile
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, searchStr+"*"))
		read = ReplaceString("\t", ReplaceString(" ", read, ""), "")
		read = StringByKey(ReplaceString(" ", searchStr, ""),read,"=","\r")
		Close fileID
	endif
	return menuStr+SelectString(strlen(read),"","/"+read)
End

//################################################################################################

Function QuickMenuWrapper(dim, type)
	Variable dim
	String type
	
	String dataList = ""
	Variable i=0
	do
		String name = GetBrowserSelection(i)
		if (Exists(name) == 1)
			dataList += name + ";"
		endif
		i += 1
	while(strlen(name) > 0)
	
	StrSwitch (type)
		case "QProfile":
			Variable mode=1, P1 = 0, P2 = inf
			Variable limitRange = (GetkeyState(0) & 2^2) != 0
			if (limitRange)																		// when ctrl / command is pressed ask for data range
				Prompt mode, "the range is in ...", popup "scaled values;data points;"
				Prompt P1, "from:"
				Prompt P2, "to:"
				DoPrompt "Integration Over a Limited Range.", mode, P1, P2
				if (V_flag)
					return -1
				endif
			endif
			
			for (i=0; i<ItemsInList(dataList); i+=1)
				if (limitRange)
					if(mode==1)
						QProfile($StringFromList(i,dataList),dim,from=P1,to=P2)
					else
						QProfile($StringFromList(i,dataList),dim,fromPoint=P1,toPoint=P2)
					endif
				else
					QProfile($StringFromList(i,dataList),dim)
				endif
			endfor
		break
		case "QImageSplit":
			for (i=0; i<ItemsInList(dataList); i+=1)
				QImageSplit($StringFromList(i,dataList),dim)
			endfor
		break
	EndSwitch
	return 0
End

//________________________________________________________________________________________________
//	Quick Profile - averages 2D/3D waves in one direction (along x, y or z) for data reduction
//________________________________________________________________________________________________

Function QProfile(inwave, dim[,from,to,fromPoint,toPoint])
	Wave inwave
	Variable dim 		// 0 = V, 1 = H, 2 = L
	Variable from,to,fromPoint,toPoint
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2)
		Abort "This is meant for 2D or 3D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif

	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	String outName = name[0,chars] + SelectString(dim-1, "_prV", "_prH","_prL")					// the name of the output wave
	
	Variable tmpP, P1 = 0, P2 = DimSize(inwave,dim)-1
	if (!ParamIsDefault(from) || !ParamIsDefault(to) || !ParamIsDefault(fromPoint) || !ParamIsDefault(toPoint))
		if(IgorVersion()<8.0)
			Abort "Cannot use sub-range with this Igor version!"
		endif
#if(IgorVersion()>=8.0)
		P1 = !ParamIsDefault(fromPoint) ? limit(fromPoint, 0, DimSize(inwave,dim)-1) : P1
		P2 = !ParamIsDefault(toPoint)   ? limit(toPoint,   0, DimSize(inwave,dim)-1) : P2
		P1 = !ParamIsDefault(from) && numtype(from) == 0 ? limit(ScaleToIndex(inwave, from, dim ), 0, DimSize(inwave,dim)-1)  : P1
		P2 = !ParamIsDefault(to)   && numtype(to)   == 0 ? limit(ScaleToIndex(inwave, to,   dim ), 0, DimSize(inwave,dim)-1)  : P2
		if (P1 > P2)
			tmpP = P1
			P1 = P2
			P2 = tmpP
		endif
		if (dim==0)
			Duplicate/Free/RMD=[P1,P2][][] inwave, tempwave
		elseif (dim==1)
			Duplicate/Free/RMD=[][P1,P2][] inwave, tempwave
		elseif (dim==2)
			Duplicate/Free/RMD=[][][P1,P2] inwave, tempwave
		endif
		Wave inwave = tempwave
#endif
		chars = chars - strlen(num2str(P1)) - strlen(num2str(P2)) - 1
		outName = name[0,chars] + SelectString(dim-1, "_V[", "_H[", "_L[") + num2str(P1)+"-"+num2str(P2) + "]"
	endif
	
	if (dim==0)
		MatrixOP/O $outName = sumRows(inwave^t)/numRows(inwave)
	elseif (dim==1)
		MatrixOP/O $outName = sumRows(inwave)/numCols(inwave)
	elseif (dim==2)
		MatrixOP/O $outName = sumBeams(inwave)/waveLayers(inwave)
	endif
	Wave outwave = $outName
	
	if (wdim == 3 && dim < 2)																	// for profiles of 3D waves -> rotate result to 2D map
		MatrixOP/Free tempVol = transposeVol(outwave,1)
		Duplicate/O tempVol, outwave
		SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), outwave
	endif
	
	String wUnit = WaveUnits(inwave,dim)
	String notes = note(inwave) + "\r" + SelectString(dim-1, "Rows", "Columns","Layers") + " of " + name	// write a note of what has been done
	notes += " integrated from "+num2str(DimOffset(inwave,dim))+" to "+num2str(DimOffset(inwave,dim)+(DimSize(inwave,dim)-1)*DimDelta(inwave,dim))
	notes += SelectString(strlen(wUnit),""," ")+wUnit+" (points "+num2str(P1)+" to "+num2str(P2)+")"
	Note outwave, notes
	
	if (dim < 2)
		SetScale/P x, DimOffset(inwave,1-dim), DimDelta(inwave,1-dim), WaveUnits(inwave,1-dim), outwave
	elseif (dim == 2)
		CopyScales inwave, outwave
	endif
	SetScale d 0,0, WaveUnits(inwave,-1), outwave												// also copy the data units
	
	return 0
End

//________________________________________________________________________________________________
//	Image split - splits 2D/3D data into individual waves along x, y or z
//________________________________________________________________________________________________

Function QImageSplit(inwave,dim)
	Wave inwave
	Variable dim 		// 0 = V, 1 = H, 2 = L
	
	Variable wdim = WaveDims(inwave)
	String name = NameOfWave(inwave)
	
	dim = round(dim)
	if (dim < 0 || dim > 2)
		Abort "Wrong dimension parameter. Must be 0, 1 or 2"
	endif
	if (wdim < 2)
		Abort "This is meant for 2D or 3D waves!"
	endif
	if (wdim != 3 && dim == 2)
		Abort "This is meant for 3D waves!"
	endif
	
	DFREF saveDFR = GetDataFolderDFR()
	NewDataFolder/O/S $name
	
	Variable i
	String wname, unit, scaledpos
	String notes = note(inwave) + "\r"															// scale & write a note of what has been done
	Switch (dim)
		case 0:
			unit = WaveUnits(inwave,0)
			for (i = 0; i < DimSize(inwave, 0); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,0)+i*DimDelta(inwave,0)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "row_img_" + num2str(i) + scaledpos
					MatrixOP/Free tempData = row(inWave,i)
					MatrixOP/O $wname = transposeVol(tempData,4)
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "row_" + num2str(i) + scaledpos
					MatrixOP/O $wname = row(inWave,i)^t
					SetScale/P x, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
				endif
				SetScale d 0,0, WaveUnits(inwave,-1), $wname
				Note $wname, notes+"row "+num2str(i)+" of "+NameOfWave(inwave)
			endfor
		break
		case 1:
			unit = WaveUnits(inwave,1)
			for (i = 0; i < DimSize(inwave, 1); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,1)+i*DimDelta(inwave,1)) + SelectString(strlen(unit),""," ") + unit + ")"
				if (wdim == 3)
					wname = "col_img_" + num2str(i) + scaledpos
					MatrixOP/Free tempData = col(inWave,i)
					MatrixOP/O $wname = transposeVol(tempData,1)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
					SetScale/P y, DimOffset(inwave,2), DimDelta(inwave,2), WaveUnits(inwave,2), $wname
				else
					wname = "col_" + num2str(i) + scaledpos
					MatrixOP/O $wname = col(inWave,i)
					SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				endif
				SetScale d 0,0, WaveUnits(inwave,-1), $wname
				Note $wname, notes+"column "+num2str(i)+" of "+NameOfWave(inwave)
			endfor
		break
		case 2:
			unit = WaveUnits(inwave,2)
			for (i = 0; i < DimSize(inwave, 2); i += 1)
				scaledpos = " (" + num2str(DimOffset(inwave,2)+i*DimDelta(inwave,2)) + SelectString(strlen(unit),""," ") + unit + ")"
				wname = "layer_" + num2str(i) + scaledpos
				MatrixOP/O $wname=inWave[][][i]
				SetScale/P x, DimOffset(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0), $wname
				SetScale/P y, DimOffset(inwave,1), DimDelta(inwave,1), WaveUnits(inwave,1), $wname
				SetScale d 0,0, WaveUnits(inwave,-1), $wname
				Note $wname, notes+"layer "+num2str(i)+" of "+NameOfWave(inwave)
			endfor
		break
	EndSwitch
	SetDataFolder saveDFR
	
	return 0
End

//________________________________________________________________________________________________
//	Combines individual 1D or 2D datasets to one 2D/3D dataset
//________________________________________________________________________________________________

Function CombineDataSets([stack3D])
	Variable stack3D																			// stack3D = 1 concatenates the data without summing
	
	String name = "", list = "", waves= ""
	Variable i = 0, j, items
	do																							// make a list of the selected data
		name = GetBrowserSelection(i)
		if (strlen(name) <= 0)
			break
		endif
		if (Exists(name) == 1)																	// check if it's a wave
			list += name + ";"
			waves+= ParseFilePath(0, name, ":", 1, 0) + ";"										// this is just a list of input data for the wave notes
		endif
		i+=1
	while(1)
	list  = Sortlist(list, ";", 16)
	items = itemsinlist(list)
	if (!items)
		return -1
	endif
	
	// +++++++++++++++++++++++++++++++++ build an output name ++++++++++++++++++++++++++++++++++++
	String number1, number2, outname
	name = ParseFilePath(0, StringFromList(items-1, list), ":", 1, 0)							// try to get the last data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number2													// get the end number in the name
	name = ParseFilePath(0, StringFromList(0, list), ":", 1, 0)									// try to get the first data number in the list
	SplitString/E="(\d+)(?!.*\d)" name, number1
	
	outname = RemoveEnding(ReplaceString("'",name,""), number1)
	outname = RemoveEnding(outname, "_")
	if(strlen(number1) > 0 && strlen (number2) > 0)												// add range to name if both numbers are set
		outname += "_"+number1+"-"+number2
	endif
	Variable chars = IgorVersion() < 7.00 ? 26 : inf
	outname = outname[0,chars] + "_cmb"
	
	if (!ParamIsDefault(stack3D) && stack3D == 1)
		Concatenate/O list, $outname
		Wave out = $outname																		// label each layer with their input
		for (i = 0; i < items; i += 1)
			Wave work = $StringFromList(i, list)
			name = ReplaceString("'",ReplaceString("\"",ReplaceString(",",ReplaceString(";",NameOfWave(work),""),""),""),"")[0,30]
			SetDimLabel 2, i, $name, out
		endfor
		return 0																				// stop here
	endif
	
	// ++++++++++++++++++++++++++++ continue when stack3D is not set +++++++++++++++++++++++++++++
	Make/FREE/N=(items) xSizes, ySizes, Offset, Deltas, RngEnd									// temporary data holding all the scaling and unit values from all waves
	//Make/FREE/N=(items)/T xUnits, yUnits
	
	Wave work 		= $StringFromList(0, list)													// extract y-scaling and names from first wave
	Variable yOffset= DimOffset(work, 1)
	Variable yDelta	= DimDelta(work, 1)
	String dUnit	= WaveUnits(work, -1)
	String xUnit	= WaveUnits(work, 0)
	String yUnit	= SelectString(WaveDims(work)>1,"",WaveUnits(work, 1))
	String wname	= NameOfWave(work)
	String notes	= note(work)
	
	for (i = 0; i < items; i += 1)
		Wave work = $StringFromList(i, list)
		xSizes[i] = DimSize(work, 0)
		ySizes[i] = DimSize(work, 1) > 0 ? DimSize(work, 1) : 1
		Offset[i] = DimOffset(work, 0)
		Deltas[i] = DimDelta(work, 0)
		RngEnd[i] = Offset[i] + Deltas[i] * (xSizes[i]-1)
		//xUnits[i] = WaveUnits(work, 0)
	endfor
	
	Variable Delta = Deltas[0], DeltaMisMatch = 0
	if (Delta != mean(Deltas))																	// may lead to data overrun
		DoAlert 1, "Data x spacing does not match! Do you want to use interpolation? Press no to exit."
		if (V_flag == 1)
			Delta = Delta > 0 ? WaveMin(Deltas) : WaveMax(Deltas)
			DeltaMisMatch = 1
		else
			return 0
		endif
	endif
	
	Variable WaveStart	= Delta > 0? WaveMin(Offset) : WaveMax(Offset)
	Variable WaveEnd	= Delta > 0? WaveMax(RngEnd) : WaveMin(RngEnd)
	Variable xSize		= WaveMax(xSizes)
	if (DeltaMisMatch)
		xSize = abs((WaveStart-WaveEnd)/Delta)+1
	endif
	
	Make/O/D/N=(xSize, Sum(Ysizes)) $outname = 0
	Wave out = $outname
	SetScale/I x, WaveStart, WaveEnd, xUnit, out
	SetScale/P y, yOffset, yDelta, yUnit, out
	SetScale d 0,0, dUnit, out
	Note out, notes + "\rCombined data of "+ReplaceString(";",RemoveEnding(waves),", ")

	Variable startpos = 0, endpos, startpnt, endpnt
	for (i = 0; i < itemsinlist(list); i += 1)
		endpos	  = startpos + ySizes[i]-1
		WaveStart = Offset[i]
		Wave work = $StringFromList(i, list)

		name = ReplaceString("'",ReplaceString("\"",ReplaceString(",",ReplaceString(";",NameOfWave(work),""),""),""),"")[0,30]
		for (j = 0; j < ySizes[i]; j += 1)
			SetDimLabel 1, startpos+j, $name, out												// label added columns with the input wave
		endfor

		startpnt  = x2pnt(out, pnt2x(work, 0))													// find the sub-range of valid x-values from work
		endpnt	  = x2pnt(out, pnt2x(work, xSizes[i]-1))
		startpnt += (pnt2x(work, 0) - pnt2x(out, startpnt)) >= Delta							// make sure the x-range is not outside the valid range
		endpnt	 -= (pnt2x(out, endpnt) - pnt2x(work, xSizes[i]-1)) >= Delta
		
		out[startpnt,endpnt][startpos,endpos] = work(x)[q-startpos] 
		startpos += ySizes[i]
	endfor
	
	return 0
End

//________________________________________________________________________________________________
//	Bin data in x, y, z
//________________________________________________________________________________________________

Function QuickDataBinWrapper()
	String name = "", list = ""
	Variable i = 0, items
	do																							// make a list of the selected data
		name = GetBrowserSelection(i)
		if (strlen(name) <= 0)
			break
		endif
		if (Exists(name) == 1)																	// check if it's a wave
			list += name + ";"
		endif
		i+=1
	while(1)
	items = itemsinlist(list)
	if (!items)
		return -1
	endif
	
	Variable binX = NumVarOrDefault("root:DataBin_X",1)
	Variable binY = NumVarOrDefault("root:DataBin_Y",1)
	Variable binZ = NumVarOrDefault("root:DataBin_Z",1)
	Prompt binX, "X dimension:"
	Prompt binY, "Y dimension:"
	Prompt binZ, "Z dimension:"
	String helptext = "Creates a lower-resolution representation of the data by using ImageInterpolate."
	DoPrompt/HELP=helptext "Enter how many points should be averaged in each dimension", binX, binY, binZ
	
	for (i = 0; i < items; i+=1)
		Wave in = $StringFromList(i, list)
		QuickBinData(in, binX, binY, binZ)
	endfor
	
	Variable/G root:DataBin_X = binX															// save for later use
	Variable/G root:DataBin_Y = binY
	Variable/G root:DataBin_Z = binZ
	return 0
End

//################################################################################################

Function QuickBinData(in, binX, binY, binZ)
	Wave in
	Variable binX, binY, binZ
	if (WaveDims(in) < 2 && IgorVersion() >= 7)													// only for images
		return 0
	endif
#if (IgorVersion() >= 7)
	binX = binX < 1 ? 1 : round(binX)
	binY = binY < 1 ? 1 : round(binY) 
	binZ = binZ < 1 ? WaveDims(in) == 3 : round(binZ)*(WaveDims(in) == 3)
	ImageInterpolate/PXSZ={binX,binY,binZ}/DEST=$(NameOfWave(in)+"_avg") Pixelate in			// average image
	Wave out = $(NameOfWave(in)+"_avg")															// transfer scales and notes
	SetScale/P x DimOffset(in,0), DimDelta(in,0)*binX, WaveUnits(in,0), out
	SetScale/P y DimOffset(in,1), DimDelta(in,1)*binY, WaveUnits(in,1), out
	SetScale/P z DimOffset(in,2), DimDelta(in,2)*binZ, WaveUnits(in,2), out
	String paraStr = ""																			// write notes
	paraStr += SelectString(binX != 1,"", "x="+num2str(binX))
	paraStr += SelectString(strlen(paraStr),"", ", ")
	paraStr += SelectString(binY != 1,"", "y="+num2str(binY))
	paraStr += SelectString(strlen(paraStr),"", ", ")
	paraStr += SelectString(binZ != 1,"", "z="+num2str(binZ))
	if (strlen(paraStr))
		Note/K out, note(in) + "\r"+NameOfWave(in)+" averaged with a bin size of "+paraStr+"."
	endif
#endif
	return 0
End

//________________________________________________________________________________________________
//	Generate profiles from 2D or 3D waves
//________________________________________________________________________________________________

static Function LoadWrapper()
	Wave/Z work = $GetBrowserSelection(0)	// first selection
	if (WaveExists(work))
		SpectraProfile(work)
	endif
	return 0
End

//################################################################################################

Function SpectraProfile(inwave)
	Wave/Z inwave
	if (!WaveExists(inwave))
		return -1
	endif
	Variable dim = WaveDims(inwave)
	if (dim < 2 || dim > 3 || WaveType(inwave,1) != 1 )
		Abort "This tool works only for numeric 2D or 3D waves."
	endif
	//+++++++++++++++++++++++++++++++++ generate working variables ++++++++++++++++++++++++++++++
	String gTitle = "Spectra Image Profile (ver. "+num2str(kVersion)+" - "+kVersionDate+"): Profile of "
	String gName = UniqueName("ProfileGeneratorGraph", 6, 1), pName = "ImageProfilePanel"
	String pPath = gName+"#"+pName
	String workF = "IP"+ReplaceString("ProfileGeneratorGraph",gName,"")+"_"+CleanupName(NameOfWave(inwave),0)[0,25]
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:$(kWorkingDir)
	NewDataFolder/O root:Packages:$(kWorkingDir):$(workF)										// the folder for temporary stuff
	DFREF stg = root:Packages:$(kWorkingDir):$(workF)
	
	String/G							stg:InwaveLocation = GetWavesDataFolder(inwave,2)		// save wave name and folder for later
	Variable/G							stg:ProfileUpdateConnector								// formula connection variable... has no purpose itself, but is an anchor for triggering updates
	Make/O/D 							stg:SectionLine											// the currently displayed profile
	Make/O/D/N=(1,DimSize(inwave,2)) 	stg:MultiSection										// for multi-profiles of 3D volumes
	Make/O/D/N=5						stg:CurrentLines = 0									// holds the range values ([0] = current working mode {0 = horizontal, 1 = vertical}, [1] = h.pos, [2] = h.width, [3] = v.pos, [4] = v.width)
	Make/O/D/N=5 						stg:LinesY, stg:LinesX									// marquee line waves for the image 
	SetScale d 0,0,WaveUnits(inwave,0), stg:LinesX												// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(inwave,1), stg:LinesY
	Duplicate/O	inwave, 				stg:StartCopy											// copy original wave for undo

	NewDataFolder/O	root:WinGlobals
	NewDataFolder/O	root:WinGlobals:$(gName)													// here all main window related data is generated
	String/G		root:WinGlobals:$(gName):S_TraceOffsetInfo = ""								// Igor defined window status parameter
	//+++++++++++++++++++++++++++++++++ generate the graph ++++++++++++++++++++++++++++++++++++++
	Wave LinesX = stg:LinesX
	Wave LinesY	= stg:LinesY
	Variable pix = 72/ScreenResolution
	Variable TFTResX= Str2Num(StringFromList(3,StringByKey("SCREEN1",IgorInfo(0),":"),","))		// get the display resolution to align window better (center of full screen)
	Variable TFTResY= Str2Num(StringFromList(4,StringByKey("SCREEN1",IgorInfo(0),":"),","))
	Variable WinTop = 70, WinLeft = 200
	Variable WinBottom	= TFTResY - 150
	Variable WinRight	= TFTResX - 350

	Display/K=2/W=(WinLeft*pix, WinTop*pix, WinRight*pix, WinBottom*pix)/N=$gName as gTitle+NameOfWave(inwave)
	AppendImage/G=1 inwave																		// don't append as RGB image
	ModifyImage ''#0 ctab= {*,*,Terrain,0}
	AppendToGraph/C=(0,0,0) LinesY vs LinesX													// add marquee lines
	ModifyGraph quickdrag(LinesY)=1,live(LinesY)=1,rgb(LinesY)=(65535,0,0)						// lines can be dragged (act on S_TraceOffsetInfo)
	
	SetWindow $gName hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes
	NVar ProfileUpdateConnector = stg:ProfileUpdateConnector									// set formula for live update
	SetFormula ProfileUpdateConnector,"LinesIntegrateUpdate(root:WinGlobals:"+gName+":S_TraceOffsetInfo)"
	//++++++++++++++++++++++++++++++++++++ now the controls ++++++++++++++++++++++++++++++++++++++
	Variable panelW = 270
	Variable panelH = 590 + 120 * (dim == 3)
	
	NewPanel/HOST=$gName/EXT=0/K=2/N=$pName/W=(0, 0, panelW, panelH) as "Spectra Profile"
	SetWindow $pPath hook(CreateProfile)=Profile_ProcessWindowEvent								// keeps track of window changes here as well (only button presses)
	SetWindow $pPath userdata(workFolder)=workF													// save work folder location
	SetWindow $gName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	TitleBox IntMode			,pos={10,10}	,size={80,23}	,title="Mode:"	 			,frame=0	,fstyle=1	,fSize=12
	CheckBox IntColCheckBox		,pos={60,10}	,size={80,23}	,title="\f04H\f00orizontal: ="		,help={"Select horizontal mode. (key: 'h')"}
	CheckBox IntRowCheckBox		,pos={165,10}	,size={80,23}	,title="\f04V\f00ertical: ||"		,help={"Select vertical mode. (key: 'v')"}
	SetVariable Position		,pos={10,33}	,size={150,23}	,title="Center:"					,help={"Center position of the selection. Move with cursor keys."}	
	SetVariable Width			,pos={10,58}	,size={150,23}	,title="Width:"						,help={"Width of the selection. This is the distance between the two lines. Move with cursor keys."}
	CheckBox FullWidth			,pos={170,34}	,size={70,23}	,title="Full Size"					,help={"Select the full range of the spectrum. (key: page up)"}
	CheckBox MinWidth			,pos={170,59}	,size={70,23}	,title="Minimal Size"				,help={"Select one slice. (key: page down)"}
	
	Button bSaveSpec			,pos={20,90}	,size={100,26}	,title="\f04S\f00ave Profile"		,help={"Save current profile. (key: 's') Press 'shift' to omit range in the wave name."}
	Button bCompare				,pos={20,125}	,size={100,26}	,title="Sna\f04p\f00shot"			,help={"Click to append and compare current profile. (key: 'p')"}
	
	CheckBox Normto1			,pos={140,90}	,size={80,23}	,title="\f04N\f00ormalize Profile"	,help={"Normalizes the profile to 1. (key: 'n')"}
	CheckBox MousetoSpec		,pos={140,112}	,size={80,23}	,title="\f04M\f00ouse Csr Profile"	,help={"Live profile under the mouse cursor. (key: 'm')"}
	CheckBox SnapToGrid			,pos={140,134}	,size={80,23}	,title="Snap Lines to \f04G\f00rid"	,help={"Snap the selection lines to full rows/columns. (key: 'g')"}

	CheckBox cPrintEdits		,pos={22,161}	,size={250,23}	,title="Print a Log of Data Edits to History"				,fstyle=2

	GroupBox NormalizeGroup		,pos={10,185}	,size={250,83}	,title="\\Zr110Normalize Image / Active Layer"	,frame=0	,fstyle=1
	CheckBox MaintainInt		,pos={22,210}	,size={80,23}	,title="Keep average Int."			,help={"For normalization: Check to maintain the (average) height of the spectrum. Otherwise the spectrum will be scaled down while normalizing."}	
	Button bNormalize			,pos={20,235}	,size={110,23}	,title="Div. by Profile"			,help={"Normalize the image or selected layer by the current profile (divide). Holding 'ctrl / command' disables smoothing (hard normalize)."}
	Button bNormUser			,pos={140,205}	,size={110,23}	,title=kUserInfoBtnName				,help={"Normalize the image or selected layer with the included measurement information, e.g., ring current."}
	Button bNormMax				,pos={140,235}	,size={110,23}	,title="Each Maximum"				,help={"Normalize the image or selected layer to the maximum of every horizontal / vertical line."}
	
	GroupBox ModifyGroup		,pos={10,275}	,size={250,235}	,title="\\Zr110Edit Image / Volume"				,frame=0	,fstyle=1
	Button bSubtractPr			,pos={20,300}	,size={130,23}	,title="Subtract Profile"			,help={"Cut the current Profile from the image or current layer. Press and hold 'ctrl / command' (or 'alt / option') to subtract along features moving with increasing (decreasing) Y values."}
	Button bRemoveSelection		,pos={20,330}	,size={130,23}	,title="Delete ROI Section"			,help={"Remove the selected data between the lines from the image or ALL layers of 3D data. (key: del)"}
	Button bShiftYAxisPlus		,pos={160,300}	,size={90,23}	,title="> Shift: +Y"				,help={"Shifts 2D / 3D data by adding Y-axis values to X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}
	Button bShiftYAxisMinus		,pos={160,330}	,size={90,23}	,title="< Shift: -Y"				,help={"Shifts 2D / 3D data by subtracting Y-axis values from X values. Hold the 'ctrl / command' key, to cut image to exclude partial data."}

	GroupBox SeparationLine1	,pos={20,359}	,size={228,1}	,frame=0
	Button bAlign				,pos={20,371}	,size={65,55}	,title="Align\nFeatures"			,help={"Use data correlation to align features in the image or current layer (press 'shift' for top to bottom alignment). Set cursors to align to a sub-region!"}
	CheckBox cAlignSmooth		,pos={95,370}	,size={80,23}	,title="Smoothing:"					,help={"Smooth profile before trying alignment (helps with low S/N)."}
	CheckBox cAlignOutput		,pos={95,412}	,size={80,23}	,title="Result Output"				,help={"Saves the used shift value for each line alignment in a wave. This can be used to log the amount of shift or analyze this result further."}
	CheckBox cAlignSubShift		,pos={185,391}	,size={80,23}	,title="Substeps"					,help={"Shifts data in sub-steps if possible (interpolation is used)."}
	CheckBox cAlignEdges		,pos={95,391}	,size={80,23}	,title="Align Edges"				,help={"Align edges instead of peak features (using the derivative)."}
	SetVariable SmoothFactor	,pos={140,368}	,size={110,23}	,title=""							,help={"The smoothing factor (no. of binomial smoothing operations) applied before alignment."}

	GroupBox SeparationLine2	,pos={20,435}	,size={228,1}	,frame=0
	Button bUndoAll				,pos={20,445}	,size={90,23}	,title="\f05U\f01ndo All"			,help={"Restore the original wave. (key: 'u')"} ,fSize=13
	Button bUndoLast			,pos={20,475}	,size={90,23}	,title="Undo Last"					,help={"Restore the data before the last action."}
	Button bSaveQuit			,pos={120,445}	,size={130,23}	,title="Accept & Quit"				,help={"Keeps any changes made to the 2D data and closes the window."}
	Button bSaveAsNew			,pos={120,475}	,size={130,23}	,title="Save as ..."				,help={"Saves the edited 2D / 3D data under a new name. The original input will be restored."}
	
	TitleBox ContrastLabel		,pos={10,522}	,size={80,23}	,title="Image Contrast:"	,frame=0	,fStyle=1
	TitleBox ContrLowLabel		,pos={10,542}	,size={24,23}	,title="Min:"				,frame=0
	TitleBox ContrHighLabel		,pos={10,562}	,size={24,23}	,title="Max:"				,frame=0
	
	PopupMenu ContrastColor		,pos={120,520}	,size={140,20}	,title=""			,value= "*COLORTABLEPOPNONAMES*"
	Slider ContrastLow			,pos={48,547}	,size={210,10}	,side=0	,vert=0	,ticks=0
	Slider ContrastHigh			,pos={48,567}	,size={210,10}	,side=0	,vert=0	,ticks=0
	
	if (dim == 3)																				// add controls for 3D data
		//GroupBox SeparationLine3,pos={15,585}	,size={238,1}	,frame=0
		GroupBox LayersGroup	,pos={10,585}	,size={250,113}	,title="\\Zr1103D Volume Controls"				,frame=0	,fstyle=1
		Button bDeleteLayer		,pos={120,608}	,size={50,23}	,title="Delete"						,help={"Deletes the selected layer from the stack. This can be undone."}
		Button bSumLayers		,pos={180,608}	,size={70,23}	,title="Merge All"					,help={"Merges layers of 3D data into one 2D image for processing. This can be undone."}
		SetVariable SetLayer	,pos={10,610}	,size={100,23}	,title="Layer:"						,help={"Selects the active layer of 3D data."}
	
		CheckBox MultiProfile	,pos={22,640}	,size={80,23}	,title="Multiple Profiles"			,help={"When activated, profiles of all other layers are displayed as well."}
		SetVariable vMulti_Dist	,pos={150,638}	,size={100,23}	,title="Separation:"				,help={"Sets the vertical separation of multiple profiles in percent of the maximum y-value."}
		
		TitleBox LayerRotMode	,pos={20,670}	,size={120,23}	,title="Rot:"				 ,frame=0
		Button bLayer_RotX_pos	,pos={48,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⤴","^")		,help={"Rotates the 3D volume around x (row) axis up."}						// U+2933
		TitleBox LayerRotX		,pos={75,670}	,size={20,23}	,title="X"		 			,frame=0	,fstyle=1
		Button bLayer_RotX_neg	,pos={86,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⤵","v")		,help={"Rotates the 3D volume around x (row) axis down."}					// U+2934
		Button bLayer_RotY_pos	,pos={119,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⤾","<")		,help={"Rotates the 3D volume around y (column) axis to the left."}			// U+293e
		TitleBox LayerRotY		,pos={146,670}	,size={20,23}	,title="Y"		 			,frame=0	,fstyle=1
		Button bLayer_RotY_neg	,pos={157,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⤿",">")		,help={"Rotates the 3D volume around y (column) axis to the right."}		// U+293f
		Button bLayer_RotZ_pos	,pos={188,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⟲","<")		,help={"Rotates the image/3D volume around z (layer) axis to the left."}	// U+27f2
		TitleBox LayerRotZ		,pos={216,670}	,size={20,23}	,title="Z"		 			,frame=0	,fstyle=1
		Button bLayer_RotZ_neg	,pos={227,665}	,size={23,23}	,title=SelectString(IgorVersion()<7,"⟳",">")		,help={"Rotates the image/3D volume around z (layer) axis to the right."}	// U+27f3

		ModifyControlList ControlNameList(pPath,";","bLayer_*") fsize=14	,fstyle=1	,disable=2*(IgorVersion() < 7.0)
		CheckBox MultiProfile														,value = 0		,proc=ModeCheckBoxControl		,disable = 2*(DimSize(inwave,2)<2 || DimSize(inwave,2)>kMaxMultiProfiles)
		SetVariable SetLayer	,bodywidth = 55	,limits={0,DimSize(inwave,2)-1,1}	,value =_NUM:0	,proc=ChangeParameterControl
		SetVariable vMulti_Dist	,bodywidth = 50	,limits={0,100,5}	,format="%g %"	,value =_NUM:10	,proc=ChangeParameterControl
	endif
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pPath,";","*") fsize=10
	endif
	
	ModifyControlList ControlNameList(pPath, ";", "b*")  proc=Profile_MainButtonProc			// define the procedure for all buttons
	Button bNormUser			,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)			// activate measurement info normalization if the conditions are right
	Button bSaveSpec			,fColor=(65535,0,0)
	CheckBox IntColCheckBox		,value = 1 ,mode = 1	,proc=ModeCheckBoxControl				// radio button checked => horizontal mode
	CheckBox IntRowCheckBox		,value = 0 ,mode = 1	,proc=ModeCheckBoxControl				// radio button unchecked
	CheckBox FullWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MinWidth			,value = 0				,proc=ModeCheckBoxControl
	CheckBox Normto1			,value = 0				,proc=ModeCheckBoxControl
	CheckBox SnapToGrid			,value = 0				,proc=ModeCheckBoxControl
	CheckBox MaintainInt		,value = 0
	CheckBox cAlignSmooth		,value = 0
	CheckBox cAlignOutput		,value = 0
	CheckBox cAlignSubShift		,value = 1
	SetVariable Position		,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable Width			,bodywidth = 110		,proc=ChangeParameterControl
	SetVariable SmoothFactor	,bodywidth = 55			,limits={0,32767,1}		,value=_NUM:(ceil(DimSize(inwave,0)*kAlignSmooth/100))
	PopupMenu ContrastColor		,bodyWidth = 140		,mode=1+WhichListItem("Terrain", CTabList())	,proc=Profile_ContrastColorProc
	
	Duplicate/Free/R=[][][0] inwave, tempLayer
	MatrixOp/free tempScale = replace(tempLayer,inf,0)											// make sure to remove infs, which mask the real maximum
	Variable wMin = WaveMin(tempScale), wMax = WaveMax(tempScale)
	Slider ContrastLow			,value=wMin				,limits={wMin,wMax,0}	,proc=Profile_ContrastSliderProc
	Slider ContrastHigh			,value=wMax				,limits={wMin,wMax,0}	,proc=Profile_ContrastSliderProc
	//+++++++++++++++++++++++++++++++++++++ set dependency +++++++++++++++++++++++++++++++++++++++
	Profile_AppendSectionTrace(pPath)															// add the ROI lines & set variable limits for the first time
	Profile_UpdateAxes(pPath, -2)																// flips the axes according to the wave's delta scaling (-2 means 'reset')
	Profile_SetMinFullLineValues(pPath,1,0,0)													// set initial values for row mode (position & width)
	Profile_SetMinFullLineValues(pPath,1,1,0)													// set initial values for col mode

	Profile_UpdateLinesAndProfile(pPath)														// first update to generate initial profile
	Profile_SetFullMinCheckbox(pPath)
	DoWindow/F $gName
	return 0
End

//################################################################################################

static Function/DF getWorkFolder(pPath)															// grab current working folder from panel info
	String pPath
	String folder = ""
	if (strlen(pPath))
		folder = GetUserData(pPath,"","workFolder")
	endif
	if (strlen(folder))
		return root:Packages:$(kWorkingDir):$(folder)
	else
		return root:Packages:$(kWorkingDir)														// backwards compatibility with older sessions
	endif
End

//################################################################################################

static Function/S getPanel(gName)
	String gName
	String subPanel = StringFromList(0,RemoveFromList("AskQuitPanel",ChildWindowList(gName)))	// grab attached panel from list
	if(strlen(subPanel))
		return gName+"#"+subPanel
	else																						// compatibility with older versions
		if (WinType("ProfileGeneratorWindow"))
			return "ProfileGeneratorWindow"
		else
			return ""
		endif
	endif
End

//################################################################################################

static Function/S getGraph(pPath)
	String pPath
	return StringFromList(0,pPath,"#")
End

//################################################################################################

static Function Profile_QuitProgram(inPath, Undo)
	String inPath
	Variable Undo
	
	String gName = getGraph(inPath)																// can be called from AskQuitPanel as well => construct panel path manually
	String pPath = getPanel(gName)
	DFREF stg = getWorkFolder(pPath)
	DFREF pkroot = getWorkFolder("")
	DFREF pkg = root:Packages:
	
	SetWindow $pPath hook(CreateProfile)=$""													// release hook to not trigger any events
	SetWindow $gName hook(CreateProfile)=$""
	KillWindow $getGraph(pPath)																	// first, close image window
	NVar/Z ProfileUpdateConnector = stg:ProfileUpdateConnector
	if (!NVAR_Exists(ProfileUpdateConnector))
		return -1																				// obviously the folder does not exist anymore
	endif
	SetFormula ProfileUpdateConnector,""														// reset formula
	
	if (Undo)																					// for some strange hook double call reason, the undo has to take place AFTER window reset
		SVAR wLoc = stg:InwaveLocation															// reference the modified wave
		Wave inwave = $wLoc
		Duplicate/O stg:StartCopy, inwave														// undo the changes before quitting
	endif

	KillDataFolder/Z root:WinGlobals:$(gName)
	if (CountObjectsDFR(root:WinGlobals, 4) == 0)												// backwards compatibility - makes sure to not kill the root folder with other open instances
		KillDataFolder/Z root:WinGlobals
	endif
	if (CountObjectsDFR(stg, 4) == 0)
		KillDataFolder/Z stg
	endif
	Variable Objects = CountObjectsDFR(pkroot, 1)+CountObjectsDFR(pkroot, 2)+CountObjectsDFR(pkroot, 3)+CountObjectsDFR(pkroot, 4)
	if (Objects == 0 && DataFolderRefStatus(pkroot) != 0)
		KillDataFolder pkroot
	endif
	Objects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
	if (Objects == 0)																			// clean up Packages folder if empty
		KillDataFolder pkg
	endif
	return 0
End

//################################################################################################

static Function Profile_AppendSectionTrace(pPath)												// appends the profile parallel to the selection lines
	String pPath
	String gName = getGraph(pPath)
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave SecL	= stg:SectionLine
	Wave/Z cmpL	= stg:CompareLine
	Wave/Z multi= stg:MultiSection
	Wave inwave	= $wLoc
	Variable i, Mode = Lines[0]
	
	Profile_SetPosWidthControls(pPath)															// initialize center and width controls
	
	String getSplit
	getSplit = GetUserData(pPath, "", "vertAxSplit")											// axes may have been dragged => get new split location
	Variable vertAx = strlen(getSplit) ? str2num(getSplit) : 0.3
	getSplit = GetUserData(pPath, "", "horiAxSplit")
	Variable horiAx = strlen(getSplit) ? str2num(getSplit) : 0.5

	//+++++++++++++++++++++++++++++++++ multi-profile preparation ++++++++++++++++++++++++++++++++
	String list = GrepList(TraceNameList(gName,";",1), "^MultiSection", 0) 						// first, remove all
	for (i = ItemsInList(list)-1; i > -1; i -=1)
		RemoveFromGraph/W=$gName/Z $StringFromList(i,list)	
	endfor
	
	ControlInfo/W=$pPath MultiProfile
	Variable doMulti = V_flag == 2 && V_disable == 0 && V_Value == 1 && WaveExists(multi)
	if (doMulti)
		Variable row, size = min(DimSize(multi,1)-1,DimSize(inwave,2)-1)						// 'delete layer' may have reduced the size
		ColorTab2Wave $kMultiProfileColor;	Wave M_colors
		for (i = 0; i < size+1; i +=1)										
			if (Mode)	// vertical
				AppendToGraph/W=$gName/VERT/B=bottom2 multi[][size-i]							// go backwards (in reverse order, so that the later traces are BEHIND the earlier ones)
			else		// horizontal
				AppendToGraph/W=$gName/L=left2 multi[][size-i]
			endif
			row = (i/size)*(DimSize(M_Colors, 0)-1)
			ModifyGraph/W=$gName rgb($("MultiSection"+SelectString(i>0,"","#"+num2str(i))))=(M_Colors[row][0], M_Colors[row][1], M_Colors[row][2])
		endfor
		KillWaves/Z M_colors
	endif
	
	//+++++++++++++++++++++++++++++++++++++++ main profile +++++++++++++++++++++++++++++++++++++++
	RemoveFromGraph/W=$gName/Z CompareLine, SectionLine, MouseProfile							// remove all traces
	if (Mode)	// vertical
		AppendToGraph/W=$gName/VERT/B=bottom2 SecL
		ModifyGraph/W=$gName axisEnab(left)={0,1},	axisEnab(bottom) ={vertAx+0.01,1}			// modify the main axes
		ModifyGraph/W=$gName freePos(bottom2)=0,	axisEnab(bottom2)={0,vertAx-0.01}			// align the new bottom axis
		ModifyGraph/W=$gName zero(bottom2)=4, lblPosMode(bottom2)=1
		//SetAxis/W=$gName/A/E=1 bottom2														// scale from zero
		if (WaveExists(cmpL))
			AppendToGraph/W=$gName/VERT/B=bottom2 cmpL
			ModifyGraph/W=$gName rgb(CompareLine)=(0,0,52224), lsize(CompareLine)=(1 + doMulti)
		endif
		ColorScale/W=$gName/C/N=ImageColorDisplay/F=0/Z=1/A=LB/X=((vertAx+0.01)*100)/Y=100.10 " "		// debug: add /B=(65535,0,0)
		ColorScale/W=$gName/C/N=ImageColorDisplay image=$StringFromList(0,ImageNameList(gName, ";"))
		ColorScale/W=$gName/C/N=ImageColorDisplay vert=0, side=2, widthPct=100-(vertAx+0.01)*100, height=8, tickLen=5
		ColorScale/W=$gName/C/N=ImageColorDisplay tickLblRot=0, fsize=10, lblMargin=0, minor=1, tickUnit=1
		
		ModifyGraph/W=$gName mirror=1
		ModifyGraph/W=$gName margin(top)=40, margin(right)=0, mirror(bottom)=2
	else		// horizontal
		AppendToGraph/W=$gName/L=left2 SecL
		ModifyGraph/W=$gName axisEnab(bottom)={0,1}, axisEnab(left) ={horiAx+0.01,1}
		ModifyGraph/W=$gName freePos(left2)=0,		 axisEnab(left2)={0,horiAx-0.01}
		ModifyGraph/W=$gName zero(left2)=4, lblPosMode(left2)=1
		//SetAxis/W=$gName/A/E=1 left2
		if (WaveExists(cmpL))
			AppendToGraph/W=$gName/L=left2 cmpL
			ModifyGraph/W=$gName rgb(CompareLine)=(0,0,52224), lsize(CompareLine)=(1 + doMulti)
		endif
		ColorScale/W=$gName/C/N=ImageColorDisplay/F=0/Z=1/A=LT/X=100.10/Y=0.00 " "				// debug: add /B=(65535,0,0)
		ColorScale/W=$gName/C/N=ImageColorDisplay image=$StringFromList(0,ImageNameList(gName, ";"))
		ColorScale/W=$gName/C/N=ImageColorDisplay vert=1, side=1, heightPct=100-(horiAx+0.01)*100, width=8, tickLen=5
		ColorScale/W=$gName/C/N=ImageColorDisplay tickLblRot=90, fsize=10, lblMargin=0, minor=1, tickUnit=1
		
		ModifyGraph/W=$gName mirror=1
		ModifyGraph/W=$gName margin(top)=0, margin(right)=40, mirror(left)=2
	endif
	ModifyGraph/W=$gName highTrip=100000, standoff=0, minor=1
	ModifyGraph/W=$gName ZisZ=1, notation=1, rgb(SectionLine)=(65280,0,0)
	ModifyGraph/W=$gName lsize(SectionLine)=(1 + doMulti)
	
	if (doMulti)																				// also add color indicator for multi-profiles
		if (Mode)
			ModifyGraph mirror(bottom2)=2
			ColorScale/W=$gName/C/N=ProfileColorDisplay/F=0/Z=1/A=LB/X=0.00/Y=100.10
			ColorScale/W=$gName/C/N=ProfileColorDisplay ctab={0,size,$kMultiProfileColor,1} ,nticks=size
			ColorScale/W=$gName/C/N=ProfileColorDisplay vert=0, side=2, widthPct=(vertAx-0.01)*100, height=8, tickLen=5, fsize=10, lblMargin=0
		else
			ModifyGraph mirror(left2)=2
			ColorScale/W=$gName/C/N=ProfileColorDisplay/F=0/Z=1/A=LB/X=100.10/Y=0.00
			ColorScale/W=$gName/C/N=ProfileColorDisplay ctab={0,size,$kMultiProfileColor,1} ,nticks=size
			ColorScale/W=$gName/C/N=ProfileColorDisplay vert=1, side=1, heightPct=(horiAx-0.01)*100, width=8, tickLen=5, fsize=10, lblMargin=0
		endif
	else
		ColorScale/K/N=ProfileColorDisplay
	endif
	
	return 0
End

//################################################################################################

static Function Profile_ShiftMultiProfile(pPath)												// shifts all profiles to to their appropriate places
	String pPath
	String gName = getGraph(pPath)
	
	DFREF stg = getWorkFolder(pPath)
	Wave/Z multi = stg:MultiSection
	
	ControlInfo/W=$pPath MultiProfile															// check for the existence of these controls
	if (V_flag != 2 || V_disable != 0 || V_Value != 1 || !WaveExists(multi))
		return 0
	endif
	
	ControlInfo/W=$pPath SetLayer;		Variable layer = V_Value
	ControlInfo/W=$pPath vMulti_Dist;	Variable offset = V_Value/100*(WaveMax(multi)-WaveMin(multi))	// trace offsets are in percent of the total range
	
	String list = GrepList(TraceNameList(gName,";",1), "^MultiSection", 0)
	Variable i, items = ItemsInList(list)
	for (i = 0; i < items; i += 1)
		ModifyGraph/W=$gName offset($StringFromList(items-1 - i,list))={,i*offset}				// offset all multi-profiles
		ModifyGraph/W=$gName  lsize($StringFromList(items-1 - i,list))=!(i==layer)
	endfor
	ModifyGraph/W=$gName	 offset(SectionLine)={,layer*offset}
	if (FindListItem("CompareLine", TraceNameList(gName,";",1)) > -1)
		ModifyGraph/W=$gName offset(CompareLine)={,layer*offset}
	endif
	
	return 0
End

//################################################################################################

static Function Profile_SetPosWidthControls(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	String unitStr = WaveUnits(inwave,1-Mode)
	Variable delta = abs(DimDelta(inwave,1-Mode))
	Variable init  = delta*round((DimSize(inwave,1-Mode)-1) / 25)								// calculate ca. 5% of width
	init = init <= delta ? 2*delta : init														// make sure that's not too small

	SetVariable Position ,win=$pPath ,limits={-inf,inf,init}	,value=Lines[1+Mode*2]	,format="%g "+unitStr	// set new control values	
	ControlInfo/W=$pPath SnapToGrid;	init = V_Value ? init+mod(init/delta,2)*delta : init	// make sure the change is even when snapped to the grid (=> symmetric width change)
	SetVariable Width	 ,win=$pPath ,limits={delta,inf,init}	,value=Lines[2+Mode*2]	,format="%g "+unitStr
	return 0
End

//################################################################################################

Function Profile_ProcessWindowEvent(s)															// main window hook
	STRUCT WMWinHookStruct &s

	Variable HookTakeover = 0
	if (WinType(s.WinName) == 7 && s.EventCode != 11)											// process only key presses for the panel
		return HookTakeover
	endif

	String gName = getGraph(s.winName)															// can be called from AskQuitPanel as well => construct panel path manually
	String pPath = getPanel(gName)
	if (!strlen(pPath))																			// no panel found
		return 0
	endif

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	if (s.EventCode > 2 && s.EventCode < 6 && CmpStr(gName,s.winName) == 0)						// for mouse events
		Variable axDivider, dragAxis = 0														// for a dynamic divider line
		sscanf StringByKey("axisEnab(x)", AxisInfo(gName, SelectString(Mode,"left","bottom")), "="), "{%f,", axDivider
		axDivider -= 0.01																		// center split position
		GetWindow $gName, psizeDC
		Variable plotRange	= Mode ? v_right - v_left : v_bottom - v_top
		Variable mouseClose = Mode ? abs((s.mouseloc.h-v_left)-axDivider*plotRange) < 5 : abs((s.mouseloc.v-v_top)-(1-axDivider)*plotRange) < 5
		Variable rMousePos	= Mode ? (s.mouseloc.h-v_left)/plotRange : 1-(s.mouseloc.v-v_top)/plotRange		// relative mouse position on the plot
	endif
	
	Switch (s.EventCode)
		case 17:	// window close button
			HookTakeover = 1
			if (!EqualWaves(stg:startCopy, inwave,-1))											// compare if the wave was changed
				if (!Wintype("#AskQuitPanel"))													// this will create a little panel in case of changes
					Variable pixConv = (CmpStr(IgorInfo(2), "Windows") == 0 && ScreenResolution == 96)? 72/ScreenResolution : 1
					Variable xCenter = (s.WinRect.right - s.WinRect.left)/2*72/ScreenResolution, yCenter = (s.WinRect.bottom - s.WinRect.top)/2*72/ScreenResolution
					NewPanel/N=AskQuitPanel/HOST=$gName/W=(xCenter-112*pixConv,yCenter-35*pixConv,xCenter+112*pixConv,yCenter+35*pixConv)
					DrawText/W=$(s.winName+"#AskQuitPanel") 40,28,"Accept Changes to Data?"
					Button bAskOK		,win=$(s.winName+"#AskQuitPanel")	,pos={12,33}	,size={60,23}		,title="Yes"	,proc=Profile_AskButton
					Button bAskUndo		,win=$(s.winName+"#AskQuitPanel")	,pos={82,33}	,size={60,23}		,title="Undo"	,proc=Profile_AskButton
					Button bAskCancel	,win=$(s.winName+"#AskQuitPanel")	,pos={152,33}	,size={60,23}		,title="Cancel"	,proc=Profile_AskButton
					if (CmpStr(IgorInfo(2), "Macintosh") == 0)
						ModifyControlList ControlNameList(s.winName+"#AskQuitPanel",";","*") fsize=10
					endif
					SetActiveSubwindow $gName													// restore focus to main window
				endif
			else
				 Profile_QuitProgram(pPath,0)													// else just quit (no undo)
			endif
		break
		case 11:	// key press
			HookTakeover = 1
			//++++++++++++++++++++++++++++++++ execute button keys ++++++++++++++++++++++++++++++
			String Key = ""
			Switch(s.keycode)
				case 112:	// p key
					Key = "bCompare"
					break
				case 115:	// s key
					Key = "bSaveSpec"
					break
				case 117:	// u key
					Key = "bUndoAll"
					break
				case 127:	// del key
					Key = "bRemoveSelection"
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteButtonCommand(pPath, Key)
				break
			endif
			//++++++++++++++++++++++++++++++ execute checkbox keys ++++++++++++++++++++++++++++++
			Switch(s.keycode)
				case 11:	// page up
					Key = "FullWidth"
					break
				case 12:	// page down
					Key = "MinWidth"
					break
				case 103:	// g key
					Key = "toggleSnapToGrid"
					break
				case 104:	// h key
					Key = "IntColCheckBox"
					CheckBox IntColCheckBox ,value = 1 ,win=$pPath
					DrawAction/W=$gName getgroup=axisSplitLine, delete
					break
				case 109:	// m key
					Key = "toggleMousetoSpec"
					break
				case 110:	// n key
					Key = "toggleNormto1"
					break
				case 118:	// v key
					Key = "IntRowCheckBox"
					CheckBox IntRowCheckBox ,value = 1 ,win=$pPath
					DrawAction/W=$gName getgroup=axisSplitLine, delete
					break
			EndSwitch
			if (StrLen(Key) > 0)
				Profile_ExecuteCheckboxAction(pPath, Key,1)
				break
			endif
			//+++++++++++++++++++++++++++++++ execute arrow keys ++++++++++++++++++++++++++++++++
			ControlInfo/W=$pPath SnapToGrid
			Variable Shift = 0, Expand = 0, snapLines = V_Value
			Switch (Mode)																		// define a shift and expand direction (positive or negative) depending on mode and pressed key
				case 0:
					Shift	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Shift )
					Expand	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Expand )
					break
				case 1:
					Shift	= s.keycode == 29 ? 1 : ( s.keycode == 28 ? -1 : Shift )
					Expand	= s.keycode == 30 ? 1 : ( s.keycode == 31 ? -1 : Expand )
					break
			EndSwitch
			Variable delta = DimDelta(inwave,1-Mode)
			Variable boost = ceil(DimSize(inwave,1-Mode)/100)									// move further for bigger waves
			Variable speed = 1																	// add acceleration modifier depending on the pressed modifier key
			speed = (s.eventMod & 2^1) != 0 ? 4 : speed
			speed = (s.eventMod & 2^3) != 0 ? 0.5 : speed
			speed = ((s.eventMod & 2^3) && (s.eventMod & 2^1)) != 0 ? 1 : speed					// move exactly one delta unit with both shift and ctrl pressed
			boost = ((s.eventMod & 2^3) && (s.eventMod & 2^1)) != 0 ? 1 : boost
			
			Lines[1 + Mode*2] += Shift * delta * speed * boost									// do the shift
			Lines[2 + Mode*2] += Expand* (1+snapLines) * abs(delta) * speed * boost				// do the expand (even expansion for snapped mode)
			Lines[2 + Mode*2]  = Lines[2 + Mode*2] < delta ? delta : Lines[2 + Mode*2]			// check for the minimal width (will normally be done by the variable control)
			
			Profile_CheckBorders(pPath)															// check if the lines go over the borders of the wave
			Profile_SetFullMinCheckbox(pPath)													// toggle check-boxes
		break
		case 3: 	// mousedown
			if (mouseClose)																		// initialize axis drag state
				SetWindow $pPath userdata(axdragOn)="1"
				HookTakeover = 1
			endif
		break
		case 5: 	// mouseup
			SetWindow $pPath userdata(axdragOn)="0"												// release axis drag
		break
		case 4:		// mouse movement
			//++++++++++++++++++++++++++++++++ axis quick drag ++++++++++++++++++++++++++++++++++
			if (!WinType(pPath))																// catches slow window buildup in Igor 6
				break
			endif
			dragAxis = str2num(GetUserData(pPath, "", "axdragOn"))
			dragAxis = numtype(dragAxis) != 0 ? 0 : dragAxis
			if (dragAxis)
                GetMouse/W=$gName
                if (!(V_flag & 1))																// mouse button released? => maybe we missed a mouse-up event?
                    SetWindow $pPath userdata(axdragOn)="0"
                    dragAxis = 0
                endif
            endif
			DrawAction/W=$gName getgroup=axisSplitLine, delete									// draws and sets divider line for dynamic axis drag
			if (dragAxis)
				axDivider = limit(rMousePos, 0.15, 0.85)										// set new divider
				ModifyGraph/W=$gName axisEnab($SelectString(Mode,"left","bottom"))	 = {axDivider + 0.01, 1}
				ModifyGraph/W=$gName axisEnab($SelectString(Mode,"left2","bottom2")) = {0, axDivider - 0.01}
				SetWindow $pPath userdata($SelectString(Mode,"horiAxSplit","vertAxSplit"))=num2str(axDivider)	// save current position into panel
				
				if (FindListItem("ImageColorDisplay", AnnotationList(gName)) > -1)				// scale image color legend
					if (Mode)
						ColorScale/W=$gName/C/N=ImageColorDisplay/X=((axDivider+0.01)*100) widthPct=100-(axDivider+0.01)*100
					else
						ColorScale/W=$gName/C/N=ImageColorDisplay heightPct=100-(axDivider+0.01)*100
					endif
				endif
				if (FindListItem("ProfileColorDisplay", AnnotationList(gName)) > -1)			// scale profile color legend as well
					if (Mode)
						ColorScale/W=$gName/C/N=ProfileColorDisplay widthPct=(axDivider-0.01)*100
					else
						ColorScale/W=$gName/C/N=ProfileColorDisplay heightPct=(axDivider-0.01)*100
					endif
				endif
				HookTakeover = 1
			endif
			if (dragAxis || mouseClose)															// draw a line for visualization
				SetDrawEnv/W=$gName linefgc=(36000,15000,59000) ,dash=2 ,linethick= 1.50
				SetDrawEnv/W=$gName gstart, gname=axisSplitLine
				if (Mode)
					DrawLine/W=$gName axDivider, -0.1, axDivider, 1.1
				else
					DrawLine/W=$gName -0.1, 1-axDivider, 1.1, 1-axDivider
				endif
				SetDrawEnv/W=$gName gstop, gname=axisSplitLine
				s.cursorCode = Mode ? 5 : 6														// change mouse cursor
				s.doSetCursor = 1
			endif
			//++++++++++++++++++++++++++++++ quick mouse profile ++++++++++++++++++++++++++++++++
			ControlInfo/W=$pPath MousetoSpec
			if (V_Value)
				ControlInfo/W=$pPath SetLayer
				Variable layer = abs(V_flag) == 5 ? V_Value : 0									// get the current layer
				
				Variable Mx = round ((AxisValFromPixel("", "Bottom", s.mouseLoc.h)-DimOffset(inwave, 0)) / DimDelta(inwave, 0))
				Variable My = round ((AxisValFromPixel("", "Left", s.mouseLoc.v)-DimOffset(inwave, 1)) / DimDelta(inwave, 1))
				if (Mx >= DimSize(inwave, 0) || My >= DimSize(inwave, 1) || Mx < 0  || My < 0)
					break
				endif
				
				RemoveFromGraph/W=$gName/Z MouseProfile
				if (Mode)																		// append one row / column under the current mouse position
					AppendToGraph/W=$gName/VERT/B=bottom2 inwave[Mx][][layer]/TN=MouseProfile
				else
					AppendToGraph/W=$gName/L=left2 inwave[][My][layer]/TN=MouseProfile
				endif
				
				Variable wmin = 0, wmax = 1
				ControlInfo/W=$pPath Normto1													// also normalize the mouse profile
				if (V_Value)
					if (Mode)
						Duplicate/Free/R=[Mx][][layer] inwave, tempProfile
					else
						Duplicate/Free/R=[][My][layer] inwave, tempProfile
					endif
					MatrixOp/free tempScale = replace(tempProfile,inf,0)						// make sure to remove infs
					wmin = wavemin(tempScale); wmax = wavemax(tempScale)
				endif
				ModifyGraph/W=$gName rgb(MouseProfile)=(0,39321,0)
				sscanf StringByKey("offset(x)", TraceInfo(gName,"SectionLine",0), "="), "{%f,%f}", Mx, My	// bring the mouse profile to the same height as the main profile
				ModifyGraph offset(MouseProfile)={,My-wmin/(wmax-wmin)} ,muloffset(MouseProfile)={,1/(wmax-wmin)}
			else
				RemoveFromGraph/W=$gName/Z MouseProfile
			endif
		break
	EndSwitch
	
	return HookTakeover
End
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
Function Profile_AskButton(s) : ButtonControl													// execute button commands of the quit sub-panel
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Strswitch (s.ctrlName)
			case "bAskUndo":
				Profile_QuitProgram(s.win, 1)
				break
			case "bAskOK":
				Profile_QuitProgram(s.win, 0)
				break
			case "bAskCancel":																	// just remove the panel
				KillWindow #AskQuitPanel
				break
		EndSwitch
	endif
	return 0
End
	
//################################################################################################

static Function Profile_UpdateLinesAndProfile(pPath)											// generates marquee lines
	String pPath
	DFREF stg = getWorkFolder(pPath)
	Wave LinesX	= stg:LinesX
	Wave LinesY	= stg:LinesY
	Wave Lines	= stg:CurrentLines
	
	Variable plus, minus
	if (Lines[0] == 0)																			// depending on mode switch line direction
		plus	= Lines[1] + Lines[2]/2
		minus	= Lines[1] - Lines[2]/2
		LinesX	= {-INF,INF,NaN,-INF,INF}
		LinesY	= {plus,plus,NaN,minus,minus}
	else
		plus	= Lines[3] + Lines[4]/2
		minus	= Lines[3] - Lines[4]/2
		LinesX	= {plus,plus,NaN,minus,minus}
		LinesY	= {-INF,INF,NaN,-INF,INF}
	endif

	ModifyGraph/W=$getGraph(pPath) offset(LinesY)={0,0}											// reset offset, trigger S_TraceOffset
	return 0
End

//################################################################################################

Function LinesIntegrateUpdate(Info)																// will be triggered when S_TraceOffset changes (connected via formula)
	String Info
	if( StrSearch(Info,"TNAME:LinesY;",0) <=0 )													// contain the Info line data?
		return 0
	endif
	String gName = StringByKey("GRAPH",Info)
	String pPath = getPanel(gName)
	
	DFREF stg = getWorkFolder(pPath)
	if (DataFolderRefStatus(stg) == 0)															// something is wrong if the data is not found => exit
		return -1
	endif
	
	SVAR wLoc	= stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	Wave/Z multi= stg:MultiSection
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	if (Mode == 0)																				// depending on mode (in row [0]) read the line data (X or Y direction)
		Lines[1] += NumberByKey("YOFFSET",Info)
	else
		Lines[3] += NumberByKey("XOFFSET",Info)
	endif

	Profile_CheckBorders(pPath)																	// check if the lines go over the borders of the wave before calculating something
	//++++++++++++++++++++++++++++++++++++ prepare integration ++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pPath SetLayer
	Variable layer	= abs(V_flag) == 5 ? V_Value : 0
	Variable center	= Lines[1 + Mode * 2]
	Variable width	= Lines[2 + Mode * 2]
	Variable delta	= DimDelta(inwave, 1-Mode)
	Variable offset	= DimOffset(inwave, 1-Mode)
	Variable size	= DimSize(inwave, 1-Mode)
	ControlInfo/W=$pPath MultiProfile
	Variable doMulti = V_flag == 2 && V_Value == 1 && DimSize(inwave, 2) > 1 && WaveExists(multi)

	Variable start	= (center -	sign(delta) * width/2 - offset)/delta + 0.5						// recalculate to rows / columns x2pnt()-like; the sign(delta) reverses the direction for negative delta values
	Variable finish	= (center +	sign(delta) * width/2 - offset)/delta + 0.5						// shift the values by 0.5 to get the full rows /columns
	//start	= round(start/delta*1000)*delta/1000												// round the values depending on the delta size (to get rid of the floating point errors)
	//finish= round(finish/delta*1000)*delta/1000
	start	= start < 0 ? 0 : start																// validate values
	finish	= finish > size ? size : finish
	//Print/D "Start initial:",start,"Finish initial:",finish									// debug

	String NoteText = ""
	NoteText += note(inwave) + "\r" + NameOfWave(inwave)+" integrated from "+num2str(ChangePrecision(center-width/2,4))
	NoteText += " to "+num2str(ChangePrecision(center+width/2,4))+" "+WaveUnits(inwave,1-Mode)+ " (width = " +num2str(width)+" "+WaveUnits(inwave,1-Mode)+")"
	if (DimSize(inwave, 2) > 1)
		NoteText += " of layer "+num2str(layer)
	endif
	
	Variable i, overhead = 0, rest = 0
	overhead = ceil(start) - start																// fraction at the beginning
	start	 = ceil(start)																		// beginning of the full columns / rows
	rest	 = finish - trunc(finish)															// rest from the end of the range
	finish	 = floor(finish)																	// maximal value
	
	if (start > finish)																			// if range is smaller than one column
		overhead += rest - 1
		rest = 0
	endif
	width /= abs(delta)																			// normalize for scaled waves

	//Print/D "Overhead:",overhead,"Start:",start,"Finish:",finish,"Rest:",rest					// debug
	//++++++++++++++++++++++++++++++++++++ now integrate ++++++++++++++++++++++++++++++++++++++++
	if (doMulti)																				// do the summing for all layers
		Redimension/D/N=(DimSize(inwave, Mode),DimSize(inwave, 2)) multi
		multi = 0
		if (overhead > 0)
			multi += (Mode == 0 ? inwave[p][start-1][q] : inwave[start-1][p][q]) * overhead
		endif
		for (i = start; i < finish; i += 1)
			multi += (Mode == 0 ? inwave[p][i][q]		: inwave[i][p][q])
		endfor
		if (rest > 0)
			multi += (Mode == 0 ? inwave[p][finish][q]	: inwave[finish][p][q]) * rest
		endif
		if (width > 0)
			multi /= width
		endif
	else
		Redimension/D/N=(DimSize(inwave, Mode)) SecPr											// set the right dimension and scale
		SecPr = 0
		if (overhead > 0)
			SecPr += (Mode == 0 ? inwave[p][start-1][layer] : inwave[start-1][p][layer]) * overhead	// part of first line
		endif
		for (i = start; i < finish; i += 1)														// full lines
			SecPr += (Mode == 0 ? inwave[p][i][layer]		: inwave[i][p][layer])
		endfor
		if (rest > 0)
			SecPr += (Mode == 0 ? inwave[p][finish][layer]	: inwave[finish][p][layer]) * rest	// rest of the last line
		endif
		if (width > 0)
			SecPr /= width																		// normalize to the integrated width
		endif
	endif
	//+++++++++++++++++++++++++++++++++++++++ normalize profile ++++++++++++++++++++++++++++++++++	
	ControlInfo/W=$pPath Normto1
	if (V_Value)																				// normalizes the spectrum to the range 0-1
		if (doMulti)
			for (i = 0; i < DimSize(multi, 1); i += 1)
				Duplicate/Free/R=[][i] multi, tempCol
				start = waveMin(tempCol); delta = waveMax(tempCol)-start
				delta = delta == 0 ? 1 : delta
				multi[][i] = (multi-start)/delta
			endfor
		else
			start = waveMin(SecPr); delta = waveMax(SecPr)-start
			delta = delta == 0 ? 1 : delta
			MatrixOP/O SecPr=(SecPr-start)/delta
		endif
	endif
	
	if (doMulti)																				// set the scale last, after MatrixOP was executed
		SetScale/P x, DimOffset(inwave,Mode), DimDelta(inwave,Mode), WaveUnits(inwave,Mode), multi
		SetScale d,0,0, WaveUnits(inwave,-1), multi
		Note/K multi, NoteText																	// write comments
		Duplicate/O/R=[][layer] multi, SecPr													// just copy out the respective layer data
		Redimension/N=(-1,0) SecPr
	else
		SetScale/P x, DimOffset(inwave,Mode), DimDelta(inwave,Mode), WaveUnits(inwave,Mode), SecPr
		SetScale d,0,0, WaveUnits(inwave,-1), SecPr												// add data units
		Note/K SecPr, NoteText
	endif
	
	Profile_UpdateLabelDisplay(pPath, gName)
	Profile_ShiftMultiProfile(pPath)															// update multiple profiles
	return 0
End

//################################################################################################

static Function Profile_UpdateLabelDisplay(pPath, gName)										// creates an annotation which shows data labels
	String pPath, gName
	if (!kShowLabelDisplay)
		return 0
	endif
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc	= stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	ControlInfo/W=$pPath SetLayer
	Variable layer	= V_Value, Mode = Lines[0]
	Variable pntPos	= round((Lines[1+Mode*2] - DimOffset(inwave,1-Mode)) / DimDelta(inwave,1-Mode))	// current center position converted to points
	String lbl = "", tmplbl = "", unit = ""
	if (DimSize(inwave,2) > 1)
		if (DimOffset(inwave,2) != 0 || DimDelta(inwave,2) != 1)								// scaled layer position
			unit = WaveUnits(inwave,2)
			lbl  = num2str(DimOffset(inwave,2)+layer*DimDelta(inwave,2))+SelectString(strlen(unit),""," "+unit)
		endif
		tmplbl = GetDimLabel(inwave, 2, layer)	 												// the layer label
	endif
	lbl   += SelectString(strlen(lbl)>0 && strlen(tmplbl)>0,"","\r") + tmplbl					// add line break
	tmplbl = GetDimLabel(inwave, 1-Mode, pntPos)												// row / column labels
	lbl   += SelectString(strlen(lbl)>0 && strlen(tmplbl)>0,"","\r") + tmplbl
	
	Variable foundTextbox = FindListItem("ImageLabelDisplay", AnnotationList(gName)) > -1
	if((strlen(lbl) && !foundTextbox) || foundTextbox)
		TextBox/C/N=ImageLabelDisplay/F=0/A=RT SelectString(strlen(lbl),"","\\Zr075")+lbl
	endif
	return 0
End

//################################################################################################

static Function Profile_CheckBorders(pPath)														// check if the lines are within the boundaries, if not restrict
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable xdel = DimDelta(inwave,0), xoff = DimOffset(inwave,0)
	Variable ydel = DimDelta(inwave,1), yoff = DimOffset(inwave,1)
	Variable rows = DimSize(inwave,0) , cols = DimSize(inwave,1)
	Variable colPos	= Lines[1], colWidth= Lines[2]
	Variable rowPos	= Lines[3], rowWidth= Lines[4]
	//++++++++++++++++++++++++++ check for alignment with data grid ++++++++++++++++++++++++++++++
	Variable halfShift = 0
	ControlInfo/W=$pPath SnapToGrid
	if (V_Value)
		rowWidth	= xdel * round(rowWidth/xdel)
		colWidth	= ydel * round(colWidth/ydel)
		halfShift	= xdel/2*!mod(rowWidth/xdel,2)												// for even widths the center needs to be shifted over by half a row/col
		rowPos		= xoff + xdel*round((rowPos - halfShift - xoff)/xdel) + halfShift
		halfShift	= ydel/2*!mod(colWidth/ydel,2)
		colPos		= yoff + ydel*round((colPos - halfShift - yoff)/ydel) + halfShift
	endif
	//++++++++++++++++++++++++++++ read and check y position data ++++++++++++++++++++++++++++++++			
	Variable Boundary1 = yoff + (colWidth*sign(ydel) - ydel)/2
	Variable Boundary2 = yoff - (colWidth*sign(ydel) - ydel)/2 + ydel*(cols-1)
	
	colPos = (colPos - Boundary1)*sign(ydel) < 0 ? Boundary1 : colPos							// too low?
	colPos = (colPos - Boundary2)*sign(ydel) > 0 ? Boundary2 : colPos							// too high?
	Lines[1] = colPos																			// return correct value
	colWidth = colWidth > abs(ydel*cols) ? abs(ydel*cols) : colWidth							// too big?
	Lines[2] = colWidth
	//++++++++++++++++++++++++++++ read and check x position data ++++++++++++++++++++++++++++++++
	Boundary1 = xoff + (rowWidth*sign(xdel) - xdel)/2
	Boundary2 = xoff - (rowWidth*sign(xdel) - xdel)/2 + xdel*(rows-1)
	
	rowPos = (rowPos - Boundary1)*sign(xdel) < 0 ? Boundary1 : rowPos							// left check
	rowPos = (rowPos - Boundary2)*sign(xdel) > 0 ? Boundary2 : rowPos							// right check
	Lines[3] = rowPos
	rowWidth = rowWidth > abs(xdel*rows) ? abs(xdel*rows) : rowWidth							// too big?
	Lines[4] = rowWidth
	
	Profile_UpdateLinesAndProfile(pPath)														// set the lines to the restricted values and recalculates result
	return 0
End

//#################################support functions for control elements#########################

Function ModeCheckBoxControl(s) : CheckBoxControl
	STRUCT WMCheckboxAction &s
	if(s.eventCode == 2)
		Profile_ExecuteCheckboxAction(s.win, s.ctrlName,s.checked)
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_ExecuteCheckboxAction(pPath, type, checked)
	String pPath, type
	Variable checked
	
	DFREF stg  = getWorkFolder(pPath)
	Wave Lines = stg:CurrentLines
	StrSwitch (type)
		case "IntColCheckBox":
			Lines[0] = 0
			RemoveFromGraph/W=$getGraph(pPath)/Z CompareLine									// remove snapshot when switching modes
			KillWaves/Z stg:CompareLine
			Profile_AppendSectionTrace(pPath)
			CheckBox IntRowCheckBox ,win=$pPath ,value = 0
		break
		case "IntRowCheckBox":
			Lines[0] = 1
			RemoveFromGraph/W=$getGraph(pPath)/Z CompareLine
			KillWaves/Z stg:CompareLine
			Profile_AppendSectionTrace(pPath)
			CheckBox IntColCheckBox ,win=$pPath ,value = 0
		break
		case "FullWidth":
			Profile_SetMinFullLineValues(pPath,1,Lines[0],checked)
		break
		case "MinWidth":
			Profile_SetMinFullLineValues(pPath,0,Lines[0],checked)
		break
		case "toggleNormto1":
			ControlInfo/W=$pPath Normto1
			CheckBox Normto1		,win=$pPath ,value = 1-V_Value
		break
		case "toggleMousetoSpec":
			ControlInfo/W=$pPath MousetoSpec
			CheckBox MousetoSpec	,win=$pPath ,value = 1-V_Value
		break
		case "toggleSnapToGrid":
			ControlInfo/W=$pPath SnapToGrid
			CheckBox SnapToGrid		,win=$pPath ,value = 1-V_Value
		case "SnapToGrid":
			Profile_SetPosWidthControls(pPath)
		break
		case "MultiProfile":
			Profile_AppendSectionTrace(pPath)
			Profile_UpdateLinesAndProfile(pPath)												// now create the profiles
		break
	EndSwitch
	
	Profile_CheckBorders(pPath)																	// check if the lines go over the borders of the wave
	Profile_SetFullMinCheckbox(pPath)															// toggle check-boxes
	return 0
End

//################################################################################################

static Function Profile_SetMinFullLineValues(pPath, type, mode, activate)	
	String pPath
	Variable type, mode, activate																// type 0: min width; type 1: full width

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable size	= DimSize(inwave,1-mode)
	Variable offset	= DimOffset(inwave,1-mode)
	Variable delta	= DimDelta(inwave,1-mode)
	
	Variable init = abs(delta)*round((size-1) / 25)												// calculate ca. 5% of width
	init = init <= abs(delta) ? 2*abs(delta) : init												// make sure that's not too small
	if (type)
		Lines[1 + mode*2]  = offset																// center position
		Lines[1 + mode*2] += size < 100 ? delta*floor((size-1)/2) : delta*(size-1)/2			// align to grid for small waves
		Lines[1 + mode*2]  = abs(Lines[1 + mode*2]) < 10^-10*init ? 0 : Lines[1 + mode*2]		// properly center around zero
		Lines[2 + mode*2]  = activate == 1 ? abs(delta)*size : init								// full with or ca. 20 %
	else
		Lines[2 + mode*2]  = activate == 1 ? abs(delta) : init									// one line or 20 %
	endif
	return 0
End

//################################################################################################

Function ChangeParameterControl(s) : SetVariableControl											// execute variable changes and sets the check-boxes
	STRUCT WMSetVariableAction &s
	if (s.eventCode == 1 || s.eventCode == 2)
		Profile_ShiftMultiProfile(s.win)														// need to shift the traces first before recalculation takes place
		if (CmpStr(s.ctrlName,"SetLayer") == 0)
			ModifyImage/W=$getGraph(s.win) ''#0 ,plane = s.dval									// replace the currently displayed layer
			Profile_UpdateImageContrast(s.win, 2)												// updates the contrast sliders for the new plane
			Profile_UpdateLinesAndProfile(s.win)												// triggers profile recalculation
		endif
		Profile_CheckBorders(s.win)																// check if the lines go over the borders of the wave
		Profile_SetFullMinCheckbox(s.win)														// set check-boxes correctly
	endif
	return 0
End
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_SetFullMinCheckbox(pPath)												// look, if a checkbox could be applied
	String pPath
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode  = Lines[0]
	Variable Delta = abs(DimDelta(inwave,1-Mode))
	CheckBox FullWidth	,win=$pPath	,value = Lines[2+Mode*2] >= Delta * DimSize(inwave,1-Mode)	// check for full width
	CheckBox MinWidth	,win=$pPath	,value = Lines[2+Mode*2] <= Delta							// check for minimal width
	return 0
End

//################################################################################################

Function Profile_ContrastSliderProc(s) : SliderControl
	STRUCT WMSliderAction &s
	if (s.eventCode & 1)
		ControlInfo/W=$s.win ContrastLow;	Variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	Variable maxVal = V_Value
		ControlInfo/W=$s.win ContrastColor;	String color = S_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$color,0}
	endif
	return 0
End

//################################################################################################

Function Profile_ContrastColorProc(s) : PopupMenuControl
	STRUCT WMPopupAction &s
	if (s.eventCode == 2)
		ControlInfo/W=$s.win ContrastLow;	Variable minVal = V_Value
		ControlInfo/W=$s.win ContrastHigh;	Variable maxVal = V_Value
		ModifyImage/W=$getGraph(s.win) ''#0 ctab={minVal,maxVal,$s.popStr,0}
	endif
	return 0
End

//################################################################################################

Function Profile_MainButtonProc(s) : ButtonControl
	STRUCT WMButtonAction &s
	if (s.eventCode == 2)
		Profile_ExecuteButtonCommand(s.win, s.ctrlName)
	endif
	return 0
End
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static Function Profile_ExecuteButtonCommand(pPath, bAction)
	String pPath, bAction
	String gName = getGraph(pPath)

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable i, Mode = Lines[0]
	
	String logStr = "", axInfo = GetUserData(gName,"","axisFlip")
	String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))		// current range
	String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))
	
	ControlInfo/W=$pPath SetLayer
	Variable layer = abs(V_flag) == 5 ? V_Value : 0
	Variable wDim  = WaveDims(inwave)
	
	Variable intensitySum = 0																	// for later re-normalization
	Variable direction = 1																		// shift / rotation direction starts positive
	Variable updateContrast = 0																	// updates the contrast sliders (1 = don't change range, 2 = reset range)
	Variable updateLayers = 0																	// updates layer-related controls
	
	Variable sX = DimSize(inwave,0)-1, sY = DimSize(inwave,1)-1, sZ = DimSize(inwave,2)-1		// values for 3D volume rotation
	Variable oX = DimOffset(inwave,0), oY = DimOffset(inwave,1), oZ = DimOffset(inwave,2)
	Variable dX = DimDelta(inwave,0), dY = DimDelta(inwave,1), dZ = DimDelta(inwave,2)
	String uX = WaveUnits(inwave,0), uY = WaveUnits(inwave,1), uZ = WaveUnits(inwave,2)
	
	Variable lastX = NumberByKey("Xaxis",axInfo);	lastX = numtype(lastX) != 0 ? 1 : lastX		// saved axis flip for 3D volume rotation
	Variable lastY = NumberByKey("Yaxis",axInfo);	lastY = numtype(lastY) != 0 ? 1 : lastY
	Variable lastZ = NumberByKey("Zaxis",axInfo);	lastZ = numtype(lastZ) != 0 ? 1 : lastZ
	Variable rotNo = NumberByKey("rotNo",axInfo);	rotNo = numtype(rotNo) != 0 ? 1 : rotNo		// number of rotations done
		
	Strswitch (bAction)
		case "bSaveQuit":
			Profile_QuitProgram(pPath, 0)
			return 0
		break
		case "bSaveSpec":
			logStr = ProfileAction_SaveProfile(pPath)
		break
		case "bSaveAsNew":
			String newName = NameOfWave(inwave)
			Prompt newName, "Enter a new name:"
			DoPrompt/Help="The data will be saved in the current folder. Any exists data with the same name will be overwritten." "Save Edits with a New Name", newName
			if (!V_flag && strlen(newName))
				if (CmpStr(GetDataFolder(1)+newName, wLoc) == 0)
					DoAlert/T="Name and Location are Identical" 0, "You cannot overwrite the original data with this functionality. Please choose a different name and / or location."
					break
				endif
				Duplicate/O	inwave, $newName /Wave=newInwave
				ReplaceWave/W=$gName image=$PossiblyQuoteName(NameOfWave(inwave)), newInwave	// replace displayed image and title
				GetWindow $gName wtitle 														// change the title
				DoWindow/T $gName,ReplaceString(NameOfWave(inwave), S_value, newName)
				Profile_UndoLastUpdate(pPath, "all", bAction)									// undo all changes to original input
				Duplicate/O	newInwave, stg:StartCopy											// now replace the start copy
				wLoc = GetWavesDataFolder(newInwave,2)											// point to the new location
				Profile_AppendSectionTrace(pPath)												// update the color scale
				logStr = "Saved under a new name: "+ newName
			endif
		break
		case "bCompare":
			Duplicate/O SecPr, stg:CompareLine													// duplicate current profile for comparison
			Profile_AppendSectionTrace(pPath)
		break
		case "bUndoAll":
			Profile_UndoLastUpdate(pPath, "all", bAction)
			logStr = "All changes undone."
			updateContrast = 2
		break
		case "bUndoLast":
			Profile_UndoLastUpdate(pPath, "last", bAction)
			logStr = "Last change undone."
			updateContrast = 2
		break
		case "bAlign":
			logStr = ProfileAction_Align(pPath)													// has its own save functionality
			updateContrast = 1
		break
		case "bDeleteLayer":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Deleted layer "+num2str(layer+1)+" of "+num2str(DimSize(inwave,2))+"."
			updateLayers = 1
			
			DeletePoints/M=2 layer,1, inwave
			if (DimSize(inwave,2) == 1)
				Redimension/N=(-1,-1,0) inwave													// convert to a proper 2D image
			endif
		break
		case "bSumLayers":																		// merges 3D layers into an 2D image
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Merged all layers into 2D image."
			updateLayers = 1
			updateContrast = 2
			
			MatrixOP/Free tMat = sumBeams(inwave)/waveLayers(inwave)							// merge layers
			CopyScales inwave, tMat;	Note tMat, note(inwave)									// copy scales and notes over
			Duplicate/O tMat, inwave
		break
		case "bNormalize":																		// normalize to profile shape
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Normalized to "+SelectString((GetkeyState(0) & 2^0),"","(exact) ") + SelectString(Mode,"horizontal","vertical")
			logStr += " profile (Range: "+from+"-"+to+")" + SelectString(wDim>2,""," of layer "+num2str(layer)) + "."
			updateContrast = 2

			MatrixOP/Free tempSec=replace(SecPr,0,inf)											// the profile should not get zero => have a big number instead to suppress this row
			if( !((GetkeyState(0) & 2^0) != 0) && kNormSmooth > 0)								// ctrl / command = no smoothing
				Smooth round(kNormSmooth), tempSec												// smooth the profile a bit
			endif
			for (i = 0; i < DimSize(inwave,1-Mode); i += 1)
				if (Mode)
					inwave[i][][layer] /= tempSec[q]
				else
					inwave[][i][layer] /= tempSec[p]	
				endif
			endfor
			intensitySum = mean(SecPr)
		break
		case "bNormMax":																		// normalize to maximum
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Normalized to each "+SelectString(Mode,"vertical","horizontal") + " line maximum" + SelectString(wDim>2,""," of layer "+num2str(layer)) + "."
			updateContrast = 2
			
			Make/FREE/D/N=(DimSize(inwave,1-Mode)) tempwave										// get a temporary wave
			for (i = 0; i < DimSize(inwave,Mode); i += 1)
				tempwave = Mode > 0 ? inwave[p][i][layer] : inwave[i][p][layer]					// extract the row / column
				Variable maxV = Wavemax(tempwave)
				maxV = maxV == 0 ? 1 : maxV
				if (Mode)
					inwave[][i][layer] /= maxV
				else
					inwave[i][][layer] /= maxV
				endif
				intensitySum += maxV															// add the values for re-normalization later
			endfor
			intensitySum /= i+1
		break
		case "bNormUser":																		// normalize to measurement value
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Normalized by "+kUserInfoBtnName+"."
			updateContrast = 2
			intensitySum = Profile_ExecuteUserInfoButton(inwave)
			Button bNormUser ,win=$pPath ,disable= 2											// disable the button to avoid double clicking (more than once is meaningless anyway)			
		break
		case "bSubtractPr":		
			Profile_UndoLastUpdate(pPath, "save", bAction)
			ProfileAction_SubtractProfile(pPath)
			logStr = SelectString(Mode,"Horizontal","Vertical") + " profile (Range: "+from+"-"+to+") subtracted" + SelectString(wDim>2,""," from layer "+num2str(layer)) + "."
			updateContrast = 1
		break
		case "bRemoveSelection":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = ProfileAction_RemoveSelection(pPath)
			updateContrast = 1
		break
		case "bShiftYAxisMinus":
			direction = -1
		case "bShiftYAxisPlus":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			Profile_ShiftButtonStatus(pPath, direction)											// enable and disable shift buttons
			ProfileAction_ShiftImage(pPath, direction)
			logStr = "Data shifted to "+SelectString(direction+1,"negative","positive")+" y-axis values."
			updateContrast = 1
		break
#if (IgorVersion() >= 7.0)
		case "bLayer_RotX_neg":
			direction = -1
		case "bLayer_RotX_pos":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Data rotated around X in "+SelectString(direction+1,"clockwise","counter-clockwise")+" direction."
			updateLayers = 1
			updateContrast = 2
			if (direction == -1)
				MatrixOP/Free tempVol = reverseRows(inwave)										// 90 deg rotation = flip and transpose
				MatrixOP/O inwave = transposeVol(tempVol,1)
			else
				MatrixOP/Free tempVol = transposeVol(inwave,1)
				MatrixOP/O inwave = reverseRows(tempVol)
			endif
			SetScale/P y, oZ + sZ*dZ*(direction == +1), sign(-direction)*dZ, uZ, inwave			// flip scales as well
			SetScale/P z, oY + sY*dY*(direction == -1), sign(+direction)*dY, uY, inwave
			Wave/Z lastcopy = stg:LastCopy
			for (i = 0; i < sY+1; i += 1)														// copy labels into the right dimension
				SetDimLabel 2, i, $GetDimLabel(lastcopy,1,((direction == +1) ? i : sY-i)), inwave
			endfor		
			for (i = 0; i < sZ+1; i += 1)
				SetDimLabel 1, i, $GetDimLabel(lastcopy,2,((direction == -1) ? i : sZ-i)), inwave
			endfor
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(lastX)+";Yaxis:"+num2str(sign(-direction)*lastZ)+";Zaxis:"+num2str(sign(+direction)*lastY)+";rotNo:"+num2str(rotNo+1)+";"
		break
		case "bLayer_RotY_neg":
			direction = -1
		case "bLayer_RotY_pos":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Data rotated around Y in "+SelectString(direction+1,"clockwise","counter-clockwise")+" direction."
			updateLayers = 1
			updateContrast = 2
			if (direction == -1)
				MatrixOP/Free tempVol = reverseCols(inwave)
				MatrixOP/O inwave = transposeVol(tempVol,3)
			else
				MatrixOP/Free tempVol = transposeVol(inwave,3)
				MatrixOP/O inwave = reverseCols(tempVol)
			endif
			SetScale/P x, oZ + sZ*dZ*(direction == +1), sign(-direction)*dZ, uZ, inwave
			SetScale/P z, oX + sX*dX*(direction == -1), sign(+direction)*dX, uX, inwave
			Wave/Z lastcopy = stg:LastCopy
			for (i = 0; i < sX+1; i += 1)
				SetDimLabel 2, i, $GetDimLabel(lastcopy,0,((direction == +1) ? i : sX-i)), inwave
			endfor		
			for (i = 0; i < sZ+1; i += 1)
				SetDimLabel 0, i, $GetDimLabel(lastcopy,2,((direction == -1) ? i : sZ-i)), inwave
			endfor
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*lastZ)+";Yaxis:"+num2str(lastY)+";Zaxis:"+num2str(sign(+direction)*lastX)+";rotNo:"+num2str(rotNo+1)+";"
		break
		case "bLayer_RotZ_neg":
			direction = -1
		case "bLayer_RotZ_pos":
			Profile_UndoLastUpdate(pPath, "save", bAction)
			logStr = "Data rotated around Z in "+SelectString(direction+1,"clockwise","counter-clockwise")+" direction."
			updateContrast = 1
			if (direction == -1)
				MatrixOP/Free tempVol = reverseCols(inwave)
				if (DimSize(inwave,2) > 1)
					MatrixOP/O inwave = transposeVol(tempVol,5)
				else
					MatrixOP/O inwave = tempVol^t												// z-axis rotation also works for 2D maps
				endif
			else
				if (DimSize(inwave,2) > 1)				
					MatrixOP/Free tempVol = transposeVol(inwave,5)
				else
					MatrixOP/Free tempVol = inwave^t
				endif
				MatrixOP/O inwave = reverseCols(tempVol)
			endif
			SetScale/P x, oY + sY*dY*(direction == +1), sign(-direction)*dY, uY, inwave
			SetScale/P y, oX + sX*dX*(direction == -1), sign(+direction)*dX, uX, inwave
			Wave/Z lastcopy = stg:LastCopy
			for (i = 0; i < sX+1; i += 1)
				SetDimLabel 1, i, $GetDimLabel(lastcopy,0,((direction == +1) ? i : sX-i)), inwave
			endfor		
			for (i = 0; i < sY+1; i += 1)
				SetDimLabel 0, i, $GetDimLabel(lastcopy,1,((direction == -1) ? i : sY-i)), inwave
			endfor
			SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*lastY)+";Yaxis:"+num2str(sign(+direction)*lastX)+";Zaxis:"+num2str(lastZ)+";rotNo:"+num2str(rotNo+1)+";"
		break
#endif
	Endswitch
	
	ControlInfo/W=$pPath MaintainInt
	if (V_Value && intensitySum != 0 && numtype(intensitySum) == 0)								// restore the average height of the spectrum (from the average height of the Profile)
		inwave[][][layer] *= intensitySum
		logStr += " Average intensity is maintained."
	endif
	
	ControlInfo/W=$pPath cPrintEdits															// print logs if desired
	if (V_flag == 2 && V_Value == 1 && strlen(logStr))
		Print NameOfWave(inwave) + " (" + time() + "): " + logStr
	endif
	
	if (updateLayers)
		Profile_UpdateLayerControls(pPath)
	endif
	Profile_UpdateAxes(pPath, StringMatch(bAction, "bLayer_*"))
	Profile_UpdateImageContrast(pPath, updateContrast)
	Profile_UpdateLinesAndProfile(pPath)
	return 0
End

//################################################################################################

static Function Profile_UpdateImageContrast(pPath, updateMode)
	String pPath
	Variable updateMode
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave inwave	= $wLoc

	ControlInfo/W=$pPath ContrastColor
	if (updateMode == 0 || abs(V_flag) != 3)
		return 0
	endif
	String color = S_Value
	
	ControlInfo/W=$pPath SetLayer
	Variable layer = abs(V_flag) == 5 ? V_Value : 0
	Variable minVal, maxVal
	
	Duplicate/Free/R=[][][layer] inwave, tempLayer
	MatrixOp/free tempScale = replace(tempLayer,inf,0)
	Variable wMin = WaveMin(tempScale), wMax = WaveMax(tempScale)
	
	if (updateMode == 1)																		// mode 1 reads current setting from the contrast sliders
		ControlInfo/W=$pPath ContrastLow
		minVal = limit(V_Value,wMin,wMax)
		ControlInfo/W=$pPath ContrastHigh
		maxVal = limit(V_Value,wMin,wMax)
		minVal = minVal >= maxVal ? wMin : minVal
	else
		minVal = wMin
		maxVal = wMax
	endif
	
	Slider ContrastLow	,win=$pPath	,value=minVal	,limits={wMin,wMax,0}
	Slider ContrastHigh	,win=$pPath	,value=maxVal	,limits={wMin,wMax,0}
	ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,maxVal,$color,0}
	
	// if (minVal == wMin && maxVal != wMax)													// this interferes with 3D data where the contrast is layer-based and is not needed
		// ModifyImage/W=$getGraph(pPath) ''#0 ctab={*,maxVal,$color,0}
	// elseif (minVal != wMin && maxVal == wMax)
		// ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,*,$color,0}
	// elseif (minVal == wMin && maxVal == wMax)
		// ModifyImage/W=$getGraph(pPath) ''#0 ctab={*,*,$color,0}
	// else
		// ModifyImage/W=$getGraph(pPath) ''#0 ctab={minVal,maxVal,$color,0}
	// endif
	return 0
End

//################################################################################################

static Function Profile_UpdateAxes(pPath, update)												// after rotation operation change axis direction and units
	String pPath
	Variable update																				// [1] = do update; [-1] = undo rotation; [-2] = reset all
	if (!update)
		return 0
	endif
	
	String gName = getGraph(pPath)
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave inwave	= $wLoc
	SetScale d 0,0,WaveUnits(inwave,0), stg:LinesX												// the data units will be displayed on the x- and y-axes
	SetScale d 0,0,WaveUnits(inwave,1), stg:LinesY
	
	Variable isR = NumberByKey("rotNo",GetUserData(gName,"","axisFlip"));	isR = numtype(isR) != 0 ? 1 : isR
	if (update == -2)																			// reset all axes to their initial orientation
		SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(DimDelta(inwave,0)))+";Yaxis:"+num2str(sign(DimDelta(inwave,1)))+";Zaxis:"+num2str(sign(DimDelta(inwave,2)))+";rotNo:0;"
	endif
	Variable doX = NumberByKey("Xaxis",GetUserData(gName,"","axisFlip"));	doX = numtype(doX) != 0 ? 1 : doX
	Variable doY = NumberByKey("Yaxis",GetUserData(gName,"","axisFlip"));	doY = numtype(doY) != 0 ? 1 : doY
	Variable doZ = NumberByKey("Zaxis",GetUserData(gName,"","axisFlip"));	doZ = numtype(doZ) != 0 ? 1 : doZ
	
	String lastCmd = GetUserData(pPath,"","lastAction")
	if (update == -1)																			// return axis direction (basically do the reverse operation)
		Variable direction = -1
		StrSwitch (lastCmd)
			case "bLayer_RotX_neg":
				direction = 1
			case "bLayer_RotX_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(doX)+";Yaxis:"+num2str(sign(-direction)*doZ)+";Zaxis:"+num2str(sign(+direction)*doY)+";rotNo:"+num2str(isR-1)+";"
			break
			case "bLayer_RotY_neg":
				direction = 1
			case "bLayer_RotY_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*doZ)+";Yaxis:"+num2str(doY)+";Zaxis:"+num2str(sign(+direction)*doX)+";rotNo:"+num2str(isR-1)+";"
			break
			case "bLayer_RotZ_neg":
				direction = 1
			case "bLayer_RotZ_pos":
				SetWindow $gName userdata(axisFlip)="Xaxis:"+num2str(sign(-direction)*doY)+";Yaxis:"+num2str(sign(+direction)*doX)+";Zaxis:"+num2str(doZ)+";rotNo:"+num2str(isR-1)+";"
			break
		EndSwitch
		doX = NumberByKey("Xaxis",GetUserData(gName,"","axisFlip"))
		doY = NumberByKey("Yaxis",GetUserData(gName,"","axisFlip"))
	endif
	
	if (doX == -1)
		SetAxis/W=$gName/A/R bottom
	else
		SetAxis/W=$gName/A bottom
	endif
	if (doY == -1)
		SetAxis/W=$gName/A/R left
	else
		SetAxis/W=$gName/A left
	endif
	
	if (update > 0 || (update == -1 && StringMatch(lastCmd,"bLayer_*")) || (update == -2 && isR))	// reset center and width to adapt to new range when the data seems to be rotated
		Profile_SetMinFullLineValues(pPath,1,0,0)
		Profile_SetMinFullLineValues(pPath,1,1,0)
		RemoveFromGraph/W=$getGraph(pPath)/Z CompareLine											// the snapshot has to go
		KillWaves/Z stg:CompareLine																	// killing the wave prevents re-appendage
	endif
	return 0
End

//################################################################################################

static Function/S ProfileAction_SaveProfile(pPath)
	String pPath

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave SecPr	= stg:SectionLine
	Wave/Z multi= stg:MultiSection
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]

	String outName, addLayer = ""
	String name = NameOfWave(inwave), unit = WaveUnits(inwave,0)
	Variable skipEnd = (GetkeyState(0) & 2^2) != 0												// enables a keyboard shortcut to just cut the name from the ending
	
	String from	= num2str(ChangePrecision(Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2,4))		// current range
	String to	= num2str(ChangePrecision(Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2,4))

	if (IgorVersion() < 7.00)
		name = name[0,27]
	endif
	
	ControlInfo/W=$pPath FullWidth
	if (V_Value || skipEnd)
		outName	= name+"_pr"+SelectString(Mode , "H", "V")
	else
		if (IgorVersion() < 7.00 && (strlen(name) + strlen(from) + strlen(to) + 5) > 31 )		// make sure to not go over the 31 character limit
			from = from[0,5]
			to	 = to[0,5]
			Variable chars = 25 - strlen(from) - strlen(to)
			outName = name[0,chars] + SelectString(Mode, "_H(" , "_V(") + from+"-"+to + ")"
		else
			outName = name + SelectString(Mode, "_H(" , "_V(") + from+" to "+to + ")"
		endif
		ControlInfo/W=$pPath SetLayer
		if (DimSize(inwave, 2) > 1 && abs(V_flag) == 5)
			sprintf addLayer, "_L%.2d", V_Value
		endif
	endif
	Duplicate/O SecPr, $(outName+addLayer)
	
	ControlInfo/W=$pPath MultiProfile															// also save the multi-profile
	if (V_flag == 2 && V_Value == 1 && DimSize(inwave, 2) > 1 && WaveExists(multi))
		Variable i, pntPos	= round((Lines[1+Mode*2] - DimOffset(inwave,1-Mode)) / DimDelta(inwave,1-Mode))		// current center position converted to points
		String dimlbl = "", lbl = "", layerunit = WaveUnits(inwave,2)							// add layer info as column labels
		for (i = 0; i < DimSize(multi, 1); i += 1)
			if (DimOffset(inwave,2) != 0 && DimDelta(inwave,2) != 1)
				lbl = num2str(DimOffset(inwave,2)+i*DimDelta(inwave,2))+SelectString(strlen(layerunit),""," "+layerunit)
			endif
			dimlbl = GetDimLabel(inwave, 2, i)
			lbl   += SelectString(strlen(lbl)>0 && strlen(dimlbl)>0,""," | ") + dimlbl
			dimlbl = GetDimLabel(inwave, 1-Mode, pntPos)										// row / column labels
			lbl   += SelectString(strlen(lbl)>0 && strlen(dimlbl)>0,""," | ") + dimlbl
			SetDimLabel 1, i, $lbl, multi
		endfor
		if (IgorVersion() < 7.00)
			outName = outName[0,27]
		endif
		Duplicate/O multi, $(outName+"_map")
	endif
	
	String logStr = "Saved "+SelectString(Mode,"horizontal","vertical")+" profile (integrated from "
	logStr += from+" "+unit+" to "+to+" "+unit+") => output to "+outName+"."
	return logStr
End
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
static Function ChangePrecision(value, numSigDigits)											// round values to select-able number of digits
	variable value, numSigDigits
	string str
	sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
End

//################################################################################################

static Function/S ProfileAction_RemoveSelection(pPath)
	String pPath
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable Mode	= Lines[0]
	Variable from	= Lines[1 + Mode * 2] - Lines[2 + Mode * 2] / 2								// calculate the position of the Lines
	Variable to		= Lines[1 + Mode * 2] + Lines[2 + Mode * 2] / 2
	Variable offset = DimOffset(inwave,1-Mode), delta = DimDelta(inwave,1-Mode)
	Variable nOff	= offset-(from-to)															// recalculate the offset (is used if the removed section is at the left / bottom side)
	//Print "start:",from,"stop:", to, "offset:", nOff											// debug (range in wave scales)
	from = round((from + delta/2 - offset)/delta)												// which points do the lines include
	to	 = round((to + delta/2 - offset)/delta)
	//Print "start:",from,"stop:", to, "count:", abs(to - from)									// debug (range in wave points)
	
	DeletePoints/M=(1-Mode) from, abs(to-from), Inwave											// delete the selected points
	if (from == 0)																				// shifts the scale only if the wave is cut at the old offset
		if (Mode)
			SetScale/P x nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		else
			SetScale/P y nOff, delta, WaveUnits(inwave,1-Mode),  inwave
		endif
	endif
	
	String logStr = SelectString(Mode,"Horizontal","Vertical")+" ROI section deleted ("
	logStr += num2str(abs(to-from))+" lines in point range "+num2str(from)+"-"+num2str(to)+")."
	return logStr
End

//################################################################################################

static Function ProfileAction_ShiftImage(pPath, shiftDir)
	String pPath
	Variable shiftDir
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave inwave = $wLoc
	Wave/Z lastcpy = stg:LastCopy																// read wave data from the copy if available
	if (!WaveExists(lastcpy))
		Wave lastcpy = $wLoc
	endif
	
	Variable modkey	= (GetKeyState(0) & 2^0) != 0												// when the 'ctrl/command' key is pressed, the excess amount will be cut from the front of the spectrum	
	Variable rows = DimSize(lastcpy,0),  cols = DimSize(lastcpy,1)
	Variable xdel = DimDelta(lastcpy,0), xoff = DimOffset(lastcpy,0)
	Variable ydel = DimDelta(lastcpy,1), yoff = DimOffset(lastcpy,1)
	
	Variable addrow = ceil((cols-1)*abs(ydel/xdel))												// how many rows need to be added for the shift
	Variable nOff	= xoff + shiftDir*yoff														// the new x-offset
	nOff += sign(ydel)<0 ? shiftDir*(cols-1)*ydel : 0											// for negative y-deltas add the full width (basically start from the bottom)
	if (modkey && (addrow < rows-10))															// delete incomplete parts
		Redimension/D/N=(rows-addrow,-1,-1) inwave
		nOff += shiftDir ==  1*sign(xdel) ? addrow*xdel : 0
	else
		Redimension/D/N=(rows+addrow,-1,-1) inwave
		nOff -= shiftDir == -1*sign(xdel) ? sign(xdel)*(cols-1)*abs(ydel) : 0					// for negative shifts the start offset needs to be shifted as well (by the y-width of the spectrum)
	endif
	SetScale/P x nOff, xdel, WaveUnits(inwave,0), inwave

	Variable i
	if (WaveDims(inwave) == 3)																	// only shift the current layer of a 3D image
		for (i = 0; i < DimSize(inwave,2); i += 1)
			Duplicate/Free/R=[][][i] lastcpy, tLayer
			Redimension/N=(-1,-1,0) tLayer
			inwave[][][i] = interp2D(tLayer, nOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)	// calculate every shifted point
		endfor
		MatrixOP/Free tMat=replaceNaNs(inwave,0)												// fill NaNs with zero
		inwave = tMat
	else
		inwave[][] = interp2D(lastcpy, nOff+p*xdel - shiftDir*(yoff+q*ydel), yoff+q*ydel)
		MatrixOP/O inwave=replaceNaNs(inwave,0)
	endif
	//++++++++++++++++++++++++++ if there are zero rows, remove them +++++++++++++++++++++++++++++
	for (i = 0; i < DimSize(inwave,0); i += 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] != 0)
			break
		endif
	endfor
	if (i > 0)
		DeletePoints 0,i, inwave
		SetScale/P x DimOffset(inwave,0)+i*DimDelta(inwave,0), DimDelta(inwave,0), WaveUnits(inwave,0),  inwave
	endif
	
	Variable excessRows = 0, lastData = -1
	for (i = DimSize(inwave,0)-1; i > 0; i -= 1)
		MatrixOp/Free partSum = Sum(Row(inwave,i))
		if (partSum[0] == 0)
			excessRows += 1
		else
			if (lastData == -1)
				lastData = i
			endif
		endif
	endfor
	if (excessRows > 0 && lastData != -1)
		DeletePoints lastData+1, excessRows, inwave
	endif
	
	return 0
End

//################################################################################################

static Function Profile_ShiftButtonStatus(pPath, mode)											// disables the button after one press to prevent multiple shifts
	String pPath
	Variable mode	// -1 = -Y; 0 = undo; 1 = +Y;
	Variable currMode = 0																		// first, find the current disable state of the two buttons
	ControlInfo/W=$pPath bShiftYAxisPlus;	currMode = (V_disable == 2) ?  1 : currMode
	ControlInfo/W=$pPath bShiftYAxisMinus;	currMode = (V_disable == 2) ? -1 : currMode
	Variable Yplus  = currMode == 0 && mode ==  1 ? 2 : 0										// deactivate the relevant button
	Variable Yminus = currMode == 0 && mode == -1 ? 2 : 0
	Button bShiftYAxisPlus  ,win=$pPath ,disable=Yplus
	Button bShiftYAxisMinus ,win=$pPath ,disable=Yminus
	Button bUndoLast		,win=$pPath ,userdata(lastShift)=num2str(mode)
	return 0
End

//################################################################################################

static Function ProfileAction_SubtractProfile(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave SecPr	= stg:SectionLine
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	
	Variable Mode = Lines[0]
	Variable xOff = DimOffset(inwave,0), xDel = DimDelta(inwave,0)
	Variable yOff = DimOffset(inwave,1), yDel = DimDelta(inwave,1)
	Variable size = DimSize(inwave,1-Mode), i
	
	ControlInfo/W=$pPath SetLayer
	Variable layer = abs(V_flag) == 5 ? V_Value : 0
	
	Variable LeftOrRight = 0
	LeftOrRight = GetKeyState(0) == 1 ?  1 : LeftOrRight										// ctrl/command for positive shift
	LeftOrRight = GetKeyState(0) == 2 ? -1 : LeftOrRight										// alt/option for negative shift
	if (Mode == 0 && LeftOrRight != 0)															// subtracts along the Y axis (increasing/decreasing with Y values)
		Variable xadd  = ceil(size*abs(yDel/xDel) )												// Add points at both sides to stay within bounds
		Variable xsize = DimSize(inwave,0)
		Variable nOff  = xOff-xDel*xadd - LeftOrRight*(Lines[1]-yOff)							// new offset: add extended range and offset from current Y-center (Lines[1])
		Make/D/Free/N=(xsize+2*xadd) tempcut
		SetScale/P x, nOff, xDel ,tempcut														// readjust scale for additional points
		tempcut[0,xadd-1]					= SecPr[0]											// fill with extended profile (endpoints will be repeated)
		tempcut[xadd,xsize+xadd-1]			= SecPr[p-xadd]
		tempcut[xsize+xadd,xsize+2*xadd-1]	= SecPr[xsize-1]
	endif

	for (i = 0; i < size; i += 1)
		if (Mode == 1)
			inwave[i][][layer] -= SecPr[q]
		else
			if(WaveExists(Tempcut))
				inwave[][i][layer] -= tempcut(x-LeftOrRight*yDel*i)								// subtract a moving profile
			else
				inwave[][i][layer] -= SecPr[p]													// subtract stationary profile
			endif
		endif
	endfor
	return 0
End

//################################################################################################

static Function/S ProfileAction_Align(pPath)
	String pPath
	
	DFREF stg = getWorkFolder(pPath)
	DFREF sav = GetDataFolderDFR()
	SVAR wLoc = stg:InwaveLocation
	Wave Lines	= stg:CurrentLines
	Wave inwave	= $wLoc
	Variable Mode = Lines[0]
	
	Wave/Z backup = stg:TempCopy																// create an extra save of the current state
	if (!WaveExists(backup))
		Duplicate inwave, stg:TempCopy
		Wave backup = stg:TempCopy
	endif
	
	String name = NameOfWave(inwave)
	Variable size = DimSize(inwave,1-Mode)
	Variable line = DimSize(inwave,Mode)

	ControlInfo/W=$pPath SetLayer;			Variable layer = abs(V_flag) == 5 ? V_Value : 0
	ControlInfo/W=$pPath cAlignSmooth;		Variable doSmth = V_Value == 1
	ControlInfo/W=$pPath cAlignOutput;		Variable output = V_Value == 1
	ControlInfo/W=$pPath cAlignEdges;		Variable edgeAlign = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pPath cAlignSubShift; 	Variable subShift = (V_flag == 2 && V_Value == 1)
	ControlInfo/W=$pPath SmoothFactor;		Variable smthFact = abs(V_Flag) == 5 ? V_Value : ceil(line*kAlignSmooth/100)

	Variable rev = (GetkeyState(0) & 2^2) != 0 ? 1 : 0											// shift = reverse sorting 0 = no, 1 = yes	
	String logStr = SelectString(subShift,"Aligned ","Subs-shift aligned ")+SelectString(edgeAlign,"","edges ")+"to the "
	
	Variable chars = IgorVersion() < 7.00 ? 25: inf
	if (output)																					// writes align shift into a file
		Make/O/D/N=(size) $(name[0,chars]+"_shft")
		Wave out = $(name[0,chars]+"_shft")
		SetScale/P x DimOffset(inwave,1-Mode), DimDelta(inwave,1-Mode), WaveUnits(inwave,1-Mode), out	
	endif

	Variable from = 0, to = line-1, count														// use the cursors to select a sub-region for alignment
	if (strlen(CsrInfo(A)) > 0)
		from = Mode ? qcsr(A,"") : pcsr(A,"")
		from = numtype(from) == 0 ? from : 0
	endif
	if (strlen(CsrInfo(B)) > 0)
		to = Mode ? qcsr(B,"") : pcsr(B,"")
		to = numtype(to) == 0 ? to : line-1
	endif
	if (from > to)
		count = from
		from  = to
		to	  = count
	endif
	
	Make/FREE/N=(line) ShiftWorkA = 0, ShiftWorkB = 0, tempWave = 0
	if (Mode)
		ShiftWorkA[from,to] = backup[0 + rev*(size-1)][p][layer]
		logStr += SelectString(rev,"left","right") + " side"
	else
		ShiftWorkA[from,to] = backup[p][0 + rev*(size-1)][layer]
		logStr += SelectString(rev,"bottom","top")
	endif
	if(doSmth && smthFact > 0)
		Smooth smthFact, ShiftWorkA
		logStr += " with smoothing (factor " + num2str(smthFact) + ")"
	endif
	if (output)
		logStr += " => output to " + NameOfWave(out)
	endif
	if (edgeAlign)																				// align the derivative instead
		Differentiate ShiftWorkA/D=tempWave
		ShiftWorkA = tempWave
	endif
	
	if (subShift)
		DebuggerOptions
		Variable sav_debug = V_debugOnError
		DebuggerOptions debugOnError=0 															// switch this off in case the fit fails
	endif
	
	SetDataFolder stg
	inwave[][][layer] = NaN
	for (count = 0; count < size; count += 1)													// find the correlation and shift for each row/column
		if (Mode)
			ShiftWorkB = backup[count][p][layer]
		else
			ShiftWorkB = backup[p][count][layer]
		endif
		if (subShift)
			Duplicate/Free ShiftWorkB, currLine
		endif
		if(doSmth && smthFact > 0)
			Smooth smthFact, ShiftWorkB
		endif
		if (edgeAlign)																			// align the derivative instead
			Differentiate ShiftWorkB/D=tempWave
			ShiftWorkB = tempWave
		endif
		if (from > 0)
			ShiftWorkB[0,from] = 0
		endif
		if (to < line-1)
			ShiftWorkB[to,line-1] = 0
		endif
		
		Variable align = SpectrumCorrelationShift(ShiftWorkA,ShiftWorkB, subShift)
		ShiftWorkB = NaN
		if (Mode)
			if (subShift)		
				ShiftWorkB[-floor(align)*(align < 0),line-1 -ceil(align)*(align >= 0)] = currLine(x+align)
				inwave[count][][layer] = ShiftWorkB[q]											// write interpolated result
			else
				inwave[count][-align*(align < 0),line-1 -align*(align >= 0)][layer] = backup[count][q+align][layer]
			endif
		else
			if (subShift)
				ShiftWorkB[-floor(align)*(align < 0),line-1 -ceil(align)*(align >= 0)] = currLine(x+align)
				inwave[][count][layer] = ShiftWorkB[p]
			else
				inwave[-align*(align < 0),line-1 -align*(align >= 0)][count][layer] = backup[p+align][count][layer]
			endif
		endif
		
		if(output)
			out[count] = DimDelta(inwave, Mode) * align
		endif
	endfor
	SetDataFolder sav
	
	if (subShift)
		DebuggerOptions debugOnError=sav_debug
	endif
	
	return logStr+"."
End

//+++++++++++++++++++++++++++++++++++++ correlation function +++++++++++++++++++++++++++++++++++++

static Function SpectrumCorrelationShift(in1,in2,sub)											// Helper function: Finds maximum correlation of two slices
	Wave in1,in2																				// data to compare
	Variable sub
	
	Duplicate/Free in2, work
	Correlate/AUTO in1, work
	WaveStats/Q work
	Variable error = 1, result = 0																// result = 0 means perfect match
	Variable optLoc = sub ? V_maxLoc : (V_maxRowLoc - x2pnt(work, 0))							// if sub-shift is active, return in scaled wave units
	if (V_sum != 0)
		if (V_maxRowLoc >= 3 && sub)
			try
				CurveFit/M=0/Q gauss, work[V_maxRowLoc-3,V_maxRowLoc+3]; AbortOnRTE
				Wave W_coef
				if (numtype(W_coef[2]) == 0 && W_coef[2] > optLoc-3 && W_Coef[2] < optLoc+3)
					result = W_coef[2]
					error = 0
				endif
			catch
				if (V_AbortCode == -4)
					Variable cfError = GetRTError(1)											// 1 to clear the error
					// Print "Error during sub-step alignment:", GetErrMessage(cfError,3)		// debug
				endif
			endtry
		endif
		if (error)
			result = optLoc
		endif
	endif
	return result
End

//################################################################################################

static Function Profile_UndoLastUpdate(pPath, mode, btn)
	String pPath, mode, btn

	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave startcopy	= stg:StartCopy
	Wave/Z lastcopy = stg:LastCopy
	Wave/Z tempcopy	= stg:TempCopy
	Wave inwave		= $wLoc

	Variable lastLayers = DimSize(inwave,2)
	StrSwitch(mode)
		case "save":
			Duplicate/O inwave, stg:LastCopy
			SetWindow $pPath userdata(lastAction)=btn											// save last button press
		break
		case "last":
			if (WaveExists(tempcopy))
				Duplicate/O tempcopy, inwave, lastcopy											// updating lastcopy will prevent loading an older version upon second press
			elseif (WaveExists(lastcopy))
				Duplicate/O lastcopy, inwave
			else
				Duplicate/O startcopy, inwave
			endif
			
			String last = GetUserData(pPath,"bUndoLast","lastShift")							// checks if shift button should be restored
			if (strlen(last) > 0)
				Profile_ShiftButtonStatus(pPath, -str2num(last))
			endif
			Profile_UpdateAxes(pPath, -1)														// reset axis orientation
			SetWindow $pPath userdata(lastAction)=btn
		break
		case "all":
			Duplicate/O startcopy, inwave
			KillWaves/Z lastcopy
			Profile_ShiftButtonStatus(pPath, 0)
			Profile_UpdateAxes(pPath, -2)														// reset axis orientation
		break
	EndSwitch
	KillWaves/Z tempcopy
	
	Button bUndoLast ,win=$pPath ,userdata(lastShift)=""										// delete shift-button info
	ControlInfo/W=$pPath bNormUser																// backwards compatibility
	if (V_flag == 1)
		Button bNormUser ,win=$pPath ,disable=2*(Profile_ActivateUserInfoButton(inwave)!=1)		// re-enable the button if data is available
	endif
	if (lastLayers != DimSize(inwave,2))														// undo has changed the number of layers
		Profile_UpdateLayerControls(pPath)
	endif
	return 0
End

//################################################################################################

static Function Profile_UpdateLayerControls(pPath)												// toggles layer-related controls
	String pPath
	DFREF stg = getWorkFolder(pPath)
	SVAR wLoc = stg:InwaveLocation
	Wave/Z multi = stg:MultiSection
	Wave inwave = $wLoc
	
	ControlInfo/W=$pPath SetLayer
	if (abs(V_flag) != 5)																		// the following controls were all introduced later
		return 0
	endif
	
	Variable layer = V_Value, lSize = DimSize(inwave,2), lLimit = limit(lsize-1,0,inf)
	ModifyImage/W=$getGraph(pPath) ''#0 ,plane = limit(layer,0,lLimit)							// select the next best layer
	SetVariable SetLayer	,win=$pPath ,disable = 2*(lsize<2) ,value=_NUM:limit(layer,0,lLimit) ,limits={0,lLimit,1}
	Button bSumLayers		,win=$pPath ,disable = 2*(lsize<2)
	
	ControlInfo/W=$pPath bDeleteLayer															// backwards compatibility - added delete button in update
	if (V_flag == 1)
		Button bDeleteLayer ,win=$pPath ,disable = 2*(lsize<2)
	endif
	
	if (IgorVersion() < 7.0)
		ModifyControlList ControlNameList(pPath,";","bLayer_*")											,win=$pPath ,disable = 2
	else
		ModifyControlList ControlNameList("",";","bLayer_RotX*")+ControlNameList("",";","bLayer_RotY*") ,win=$pPath ,disable = 2*(lsize<2)
	endif
	
	ControlInfo/W=$pPath MultiProfile															// backwards compatibility - added rotate controls and multi-layer in update	
	if (V_flag == 2)
		CheckBox MultiProfile ,win=$pPath ,disable = 2*(lsize<2 || lsize>kMaxMultiProfiles)
	endif
	
	if (WaveExists(multi))																		// make sure the multi-profile has the correct size immediately
		Redimension/D/N=(-1, lSize) multi
	endif
	
	Profile_AppendSectionTrace(pPath)															// do multi-trace updates
	return 0
End